import numpy as np

from poli.core.black_box_information import BlackBoxInformation
from poli.core.lambda_black_box import LambdaBlackBox
from poli.tests.observers.test_observers import SimpleObserver


def test_lambda_black_box_works_without_custom_info():
    def f_(x: np.ndarray) -> np.ndarray:
        return np.zeros((len(x), 1))

    f = LambdaBlackBox(
        function=f_,
    )

    assert f.info is not None
    assert (f(np.ones((10, 10))) == 0.0).all
    assert f.num_evaluations == 10


def test_lambda_black_box_with_custom_info_works():
    def f_(x: np.ndarray) -> np.ndarray:
        return np.zeros((len(x), 1))

    f = LambdaBlackBox(
        function=f_,
        info=BlackBoxInformation(
            name="zero",
            max_sequence_length=np.inf,
            aligned=False,
            fixed_length=False,
            deterministic=True,
            alphabet=None,
            discrete=True,
        ),
    )

    assert f.info is not None
    assert f.info.name == "zero"
    assert f.info.deterministic
    assert (f(np.ones((10, 10))) == 0.0).all
    assert f.num_evaluations == 10


def test_attaching_observer_to_lambda_black_box_works():
    def f_(x: np.ndarray) -> np.ndarray:
        return np.zeros((len(x), 1))

    f = LambdaBlackBox(
        function=f_,
        info=BlackBoxInformation(
            name="zero",
            max_sequence_length=np.inf,
            aligned=False,
            fixed_length=False,
            deterministic=True,
            alphabet=None,
            discrete=True,
        ),
    )

    observer = SimpleObserver()
    observer.initialize_observer(
        problem_setup_info=f.info, caller_info={"experiment_id": "example"}, seed=None
    )

    f.set_observer(observer)

    f(np.ones((10, 10)))

    assert (np.array(observer.results[0]["x"]) == 1.0).all()
    assert (np.array(observer.results[0]["y"]) == 0.0).all()


if __name__ == "__main__":
    test_attaching_observer_to_lambda_black_box_works()



from poli.core.black_box_information import BlackBoxInformation

AA = [
    "a",
    "r",
    "n",
    "d",
    "c",
    "q",
    "e",
    "g",
    "h",
    "i",
    "l",
    "k",
    "m",
    "f",
    "p",
    "s",
    "t",
    "w",
    "y",
    "v",
]

gfp_cbas_info = BlackBoxInformation(
    name="gfp_cbas",
    max_sequence_length=237,  # max len of aaSequence
    aligned=True,
    fixed_length=True,
    deterministic=False,
    alphabet=AA,
    log_transform_recommended=False,
    discrete=True,
    fidelity=None,
    padding_token="",
)



"""
This module implements Rosetta-energy Stability Predictions,
using PyRosetta-4 [1].

Rosetta is a black box that starts
with one wildtype pdb file, and assesses the
stability of a variant.
Two general running-modes are available: 
i) fast prediction with MinMover and scoring function -- set relax=False, pack=False
ii) default protocol prediction with pack and relax algorithm -- default .

[1] Chaudhury, Sidhartha, Sergey Lyskov, and Jeffrey J. Gray. 
"PyRosetta: a script-based interface for implementing molecular modeling algorithms using Rosetta." 
Bioinformatics 26.5 (2010): 689-691. https://doi.org/10.1093/bioinformatics/btq007 

"""

from __future__ import annotations

from pathlib import Path
from typing import Callable, List

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import get_inner_function
from poli.core.util.proteins.defaults import AMINO_ACIDS
from poli.objective_repository.rosetta_energy.information import (
    rosetta_energy_information,
)

CONSENT_FILE = Path(__file__).parent.resolve() / ".pyrosetta_accept.txt"


def has_opted_in(consent_file: Path = CONSENT_FILE) -> bool:
    if consent_file.exists():
        with open(consent_file, "r") as file:
            consent_status = file.read().strip()
            return consent_status == "accepted"
    return False


def opt_in_wrapper(f: Callable, *args, **kwargs):
    if not has_opted_in():
        agreement = input(
            "I have read and accept the License Agreements of PyRosetta, subject to the Rosetta™ license. ([Y]es/[N]o) \n See https://www.pyrosetta.org/home/licensing-pyrosetta and https://els2.comotion.uw.edu/product/rosetta ."
        )
        if agreement.strip().lower() == "yes" or agreement.strip().lower() == "y":
            with open(CONSENT_FILE, "w") as file:
                file.write("accepted")
            return f
        else:
            print(
                "You must accept and be in compliance with the original PyRosetta, Rosetta™ license."
            )
            raise RuntimeError
    else:
        return f


class RosettaEnergyBlackBox(AbstractBlackBox):
    """
    RosettaEnergy Black Box implementation.

    Parameters
    ----------
    wildtype_pdb_path : Path
        The path to the wildtype PDB file(s), by default None.
        NOTE: currently only for single PDB files implemented -- List of Paths not yet supported!
    score_function : str, optional
        Which Rosetta score function to use. Options are [ref2015 , default , centroid , fa, ref2015_cart, franklin2019].
        The default function references ref2015.
    seed : int, optional
        Overwrite Rosetta random seed with own integer, uses mt19937 RT reference (as per Rosetta default).
    unit : str, optional
        Output unit of black-box. Default is DDG, which is scaled difference between variant and wild-type.
        Alternatives are:
            REU -- raw energy function value,
            DREU -- energy unit delta to wild-type.
    conversion_factor : float, optional
        Scaling factor applied when computing DDGs (DDG=deltaREU/conversion_factor). Only applies to DDG conversion.
        Default is '2.9' following:
            Park, Hahnbeom, et al.
            "Simultaneous optimization of biomolecular energy functions on features from small molecules and macromolecules."
            Journal of chemical theory and computation 12.12 (2016): 6201-6212.
            DOI: https://doi.org/10.1021/acs.jctc.6b00819
    clean : bool, optional
        Flag whether to apply PDB cleaning routine prior to pose loading of reference.
        Default is True.
    relax: bool, optional
        Flag whether to apply relaxation protocol. Default enabled.
        NOTE: disable if fast compute required
    pack: bool, optional
        Flag whether to apply packing protocol. Default enabled.
        NOTE: disable if fast compute required
    cycle: int, optional
        Number of relaxation cycles applied.
    constraint_weight: int, optional
        Constraint on scoring function for atom-pair, angles, coordinates.

    Methods
    -------
    _black_box(x, context=None)
        The main black box method that performs the computation, i.e.
        it computes the stability of the mutant(s) in x.
    get_black_box_info()
        Returns BlackBox info object.

    Raises
    ------
    AssertionError
        If wildtype_pdb_path is not provided.
        If the provided energy units are not valid.
    NotImplementedError
        If energy function does not exist.

    Notes
    -----
    - The wildtype_pdb_path is a single Path object.
    """

    def __init__(
        self,
        wildtype_pdb_path: Path | List[Path],
        score_function: str = "default",
        seed: int = 0,
        unit: str = "DDG",
        conversion_factor: float = 2.9,
        clean: bool = True,
        relax: bool = True,
        pack: bool = True,
        cycle: int = 3,
        constraint_weight: float = 5.0,
        n_threads: int = 4,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        assert wildtype_pdb_path is not None
        self.force_isolation = force_isolation
        self.wildtype_pdb_path = wildtype_pdb_path
        self.score_function = score_function
        self.seed = seed
        self.unit = unit
        self.conversion_factor = conversion_factor
        self.clean = clean
        self.relax = relax
        self.pack = pack
        self.cycle = cycle
        self.constraint_weight = constraint_weight
        self.n_threads = n_threads

        inner_function = get_inner_function(
            isolated_function_name="rosetta_energy__isolated",
            class_name="RosettaEnergyIsolatedLogic",
            module_to_import="poli.objective_repository.rosetta_energy.isolated_function",
            force_isolation=self.force_isolation,
            wildtype_pdb_path=self.wildtype_pdb_path,
            score_function=self.score_function,
            seed=self.seed,
            unit=self.unit,
            conversion_factor=self.conversion_factor,
            clean=self.clean,
            relax=self.relax,
            pack=self.pack,
            cycle=self.cycle,
            constraint_weight=self.constraint_weight,
            n_threads=self.n_threads,
        )
        self.inner_function = opt_in_wrapper(inner_function)
        self.x0 = self.inner_function.x0

    def _black_box(self, x: np.ndarray, context: dict = None) -> np.ndarray:
        """
        Computes the stability of the mutant(s) in x.

        Parameters
        ----------
        x : np.ndarray
            Input array of shape [b, L] containing strings.
        context : dict, optional
            Additional context information (default is None).

        Returns
        -------
        y : np.ndarray
            The stability(DDGs/REUs/DREUs) of the mutant(s) in x.
        """
        return self.inner_function(x, context)

    def get_black_box_info(self) -> BlackBoxInformation:
        """
        Returns the black box information for Rosetta.
        """
        is_fixed_length = True
        max_sequence_length = len("".join(list(self.x0[0])))
        return BlackBoxInformation(
            name="rosetta_energy",
            max_sequence_length=max_sequence_length,
            aligned=False,
            fixed_length=is_fixed_length,
            deterministic=False,
            alphabet=AMINO_ACIDS,
            log_transform_recommended=False,
            discrete=True,
            fidelity="high",
            padding_token="",
        )


class RosettaEnergyProblemFactory(AbstractProblemFactory):
    def get_setup_information(self) -> BlackBoxInformation:
        return rosetta_energy_information

    def create(
        self,
        wildtype_pdb_path: Path | List[Path],
        score_function: str = "default",
        seed: int = 0,
        unit: str = "DDG",
        conversion_factor: float = 2.9,
        clean: bool = True,
        relax: bool = True,
        pack: bool = True,
        cycle: int = 3,
        constraint_weight: int | float = 5,
        n_threads: int = 4,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Creates a Rosetta black box instance, alongside initial
        observations.

        Parameters
        ----------
        wildtype_pdb_path : Path
            The path to the wildtype PDB file(s), by default None.
            NOTE: currently only for single PDB files implemented -- List of Paths not yet supported!
        score_function : str, optional
            Which Rosetta score function to use. Options are [ref2015 , default , centroid , fa, ref2015_cart, franklin2019].
            The default function references ref2015.
        seed : int, optional
            Overwrite Rosetta random seed with own integer, uses mt19937 RT reference (as per Rosetta default).
        unit : str, optional
            Output unit of black-box. Default is DDG, which is scaled difference between variant and wild-type.
            Alternatives are:
                REU -- raw energy function value,
                DREU -- energy unit delta to wild-type.
        conversion_factor : float, optional
            Scaling factor applied when computing DDGs (DDG=deltaREU/conversion_factor). Only applies to DDG conversion.
            Default is '2.9' following:
                Park, Hahnbeom, et al.
                "Simultaneous optimization of biomolecular energy functions on features from small molecules and macromolecules."
                Journal of chemical theory and computation 12.12 (2016): 6201-6212.
                DOI: https://doi.org/10.1021/acs.jctc.6b00819
        clean : bool, optional
            Flag whether to apply PDB cleaning routine prior to pose loading of reference.
            Default is True.
        relax: bool, optional
            Flag whether to apply relaxation protocol. Default enabled.
            NOTE: disable if fast compute required
        pack: bool, optional
            Flag whether to apply packing protocol. Default enabled.
            NOTE: disable if fast compute required
        cycle: int, optional
            Number of relaxation cycles applied.
        constraint_weight: int, optional
            Constraint on scoring function for atom-pair, angles, coordinates.
        batch_size : int, optional
            The batch size for parallel evaluation, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize evaluation, by default False.
        num_workers : int, optional
            The number of workers for parallel evaluation, by default None.
        evaluation_budget : int, optional
            The evaluation budget, by default float("inf").

        Returns
        -------
        Problem : poli.core.problem.Problem
            The poli Problem instance, containing f , x0 , y0
            where
            f : RosettaEnergyBlackBox
                The Rosetta black box instance.
            x0 : np.ndarray
                The initial observations (i.e. the wildtype as sequences
                of amino acids).
            y0 : np.ndarray
                The initial observations (i.e. the stability of the wildtype).
        """
        # Creating your black box function
        f = RosettaEnergyBlackBox(
            wildtype_pdb_path=wildtype_pdb_path,
            score_function=score_function,
            seed=seed,
            unit=unit,
            conversion_factor=conversion_factor,
            clean=clean,
            relax=relax,
            pack=pack,
            cycle=cycle,
            constraint_weight=constraint_weight,
            n_threads=n_threads,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )

        # Your first input (an np.array[str] of shape [b, L] or [b,])
        x0 = f.inner_function.x0

        return Problem(f, x0)



"""
Implements a synthetic-accessibility objective using the TDC oracles [1].

References
----------
[1] Artificial intelligence foundation for therapeutic science.
    Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022). https://doi.org/10.1038/s41589-022-01131-2
"""

from __future__ import annotations

from typing import Literal

import numpy as np
import selfies as sf

from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.chemistry.tdc_black_box import TDCBlackBox
from poli.core.chemistry.tdc_problem import TDCProblem
from poli.core.problem import Problem
from poli.core.util.chemistry.string_to_molecule import translate_smiles_to_selfies
from poli.core.util.seeding import seed_python_numpy_and_torch


class SABlackBox(TDCBlackBox):
    """Synthetic-accessibility black box implementation using the TDC oracles [1].

    Parameters
    ----------
    string_representation : Literal["SMILES", "SELFIES"], optional
        A string (either "SMILES" or "SELFIES") specifying which
        molecule representation you plan to use.
    alphabet : list[str] | None, optional
        The alphabet to be used for the SMILES or SELFIES representation.
        It is common that the alphabet depends on the dataset used, so
        it is recommended to pass it as an argument. Default is None.
    max_sequence_length : int, optional
        The maximum length of the sequence. Default is infinity.
    batch_size : int, optional
        The batch size for simultaneous execution, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize execution, by default False.
    num_workers : int, optional
        The number of workers for parallel execution, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.
    """

    def __init__(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        """
        Initialize the SABlackBox object.

        Parameters
        ----------
        string_representation : Literal["SMILES", "SELFIES"], optional
            A string (either "SMILES" or "SELFIES") specifying which
            molecule representation you plan to use.
        alphabet : list[str] | None, optional
            The alphabet to be used for the SMILES or SELFIES representation.
            It is common that the alphabet depends on the dataset used, so
            it is recommended to pass it as an argument. Default is None.
        max_sequence_length : int, optional
            The maximum length of the sequence. Default is infinity.
        batch_size : int, optional
            The batch size for parallel evaluation, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize the evaluation, by default False.
        num_workers : int, optional
            The number of workers for parallel evaluation, by default None.
        evaluation_budget : int, optional
            The maximum number of evaluations, by default float("inf").
        """
        super().__init__(
            oracle_name="SA",
            string_representation=string_representation,
            alphabet=alphabet,
            max_sequence_length=max_sequence_length,
            force_isolation=force_isolation,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="sa_tdc",
            max_sequence_length=self.max_sequence_length,
            aligned=False,
            fixed_length=False,
            deterministic=True,  # ?
            alphabet=self.alphabet,  # TODO: add alphabet once we settle for one for SMLIES/SELFIES.
            log_transform_recommended=False,
            discrete=True,
            padding_token="",
        )


class SAProblemFactory(AbstractProblemFactory):
    """Problem factory for the synthetic-accessibility problem.

    Methods
    -------
    create(...)
        Creates a synthetic-accessibility problem instance with the specified parameters.
    """

    def create(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Creates a synthetic-accessibility problem instance with the specified parameters.

        Parameters
        -----------
        string_representation : str, optional
            The string representation of the input molecules. Default is "SMILES".
        alphabet : list[str] | None, optional
            The alphabet to be used for the SMILES or SELFIES representation.
            It is common that the alphabet depends on the dataset used, so
            it is recommended to pass it as an argument. Default is None.
        max_sequence_length : int, optional
            The maximum length of the sequence. Default is infinity.
        seed:  int, optional
            The seed for random number generation. Default is None.
        batch_size:  int, optional
            The batch size for simultaneous evaluation. Default is None.
        parallelize : bool, optional
            Flag indicating whether to parallelize the evaluation. Default is False.
        num_workers:  int, optional
            The number of workers for parallel evaluation. Default is None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        force_isolation: bool, optional
            Flag indicating whether to force isolation inside the
            black box. Default is False.

        Returns
        --------
        f: SABlackBox
            The synthetic-accessibility black box function.
        x0: np.ndarray
            The initial input (taken from TDC: CC(C)(C)[C@H]1CCc2c(sc(NC(=O)COc3ccc(Cl)cc3)c2C(N)=O)C1).
        y0: np.ndarray
            The initial output (i.e. the corresponding SA).
        """
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        if string_representation.upper() not in ["SMILES", "SELFIES"]:
            raise ValueError(
                "Missing required keyword argument: string_representation: str. "
                "String representation must be either 'SMILES' or 'SELFIES'."
            )

        f = SABlackBox(
            string_representation=string_representation,
            alphabet=alphabet,
            max_sequence_length=max_sequence_length,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )

        # Initial example (from the TDC docs)
        x0_smiles = "CC(C)(C)[C@H]1CCc2c(sc(NC(=O)COc3ccc(Cl)cc3)c2C(N)=O)C1"
        x0_selfies = translate_smiles_to_selfies([x0_smiles])[0]

        # TODO: change for proper tokenization in the SMILES case.
        if string_representation.upper() == "SMILES":
            x0 = np.array([list(x0_smiles)])
        else:
            x0 = np.array([list(sf.split_selfies(x0_selfies))])

        problem = TDCProblem(
            black_box=f,
            x0=x0,
        )
        return problem



"""Tests embedding continuous toy objectives into higher dimensions.

For some objective functions that are defined in 2D
space, we give the users the affordance to embed them
in higher dimensions. This is useful for testing
higher dimensional Bayesian Optimization algorithms,
since some of them assume that the intrinsic dimensionality
of the problem is lower than the actual dimensionality.
"""

import numpy as np


def test_embed_camelback_into_high_dimensions():
    from poli import objective_factory
    from poli.objective_repository.toy_continuous_problem.register import (
        ToyContinuousProblem,
    )

    problem = objective_factory.create(
        name="toy_continuous_problem",
        function_name="camelback_2d",
        n_dimensions=2,
        embed_in=10,
    )
    f_camelback: ToyContinuousProblem = problem.black_box

    dimensions_to_embed_in = f_camelback.function.dimensions_to_embed_in

    # Testing whether the output is the same as long as we
    # are in the same subspace.
    one_x = np.random.randn(10).reshape(1, -1)
    another_x = np.random.randn(10).reshape(1, -1)

    one_x[0, dimensions_to_embed_in] = [0.0, 0.0]
    another_x[0, dimensions_to_embed_in] = [0.0, 0.0]

    assert np.allclose(
        f_camelback(one_x),
        f_camelback(another_x),
    )

    # Testing whether the output is different if we are
    # in different subspaces.
    one_x[0, dimensions_to_embed_in] = [1.0, 1.0]

    assert not np.allclose(
        f_camelback(one_x),
        f_camelback(another_x),
    )



"""Rough Mount Fuji (RMF) fitness landscapes w/ tunable ruggedness using Numpy.
See J Neidhart, IG Szendro, J Krug 
    Adaptation in Tunably Rugged Fitness Landscapes: The Rough Mount Fuji Model.
    Genetics 2014 .
    DOI: https://doi.org/10.1534/genetics.114.167668 
See Aita et al. 
    Analysis of a local fitness landscape with a model of the rough Mt. Fuji-type landscape: Application to prolyl endopeptidase and thermolysin. 
    Biopolymers 2000 .
    DOI: https://doi.org/10.1002/(SICI)1097-0282(200007)54:1<64::AID-BIP70>3.0.CO;2-R
"""



"""Tests for the budget exhaustion inside objective functions."""

import numpy as np
import pytest

from poli.core.exceptions import BudgetExhaustedException


def test_num_evaluation_tracks_correctly():
    from poli import objective_factory

    problem = objective_factory.create(
        name="toy_continuous_problem",
        function_name="ackley_function_01",
        n_dimensions=3,
        evaluation_budget=10,
    )
    f = problem.black_box

    assert f.num_evaluations == 0

    f.reset_evaluation_budget()

    x0 = np.array([[0.0, 0.0, 0.0]] * 9)

    f(x0)

    assert f.num_evaluations == 9


def test_budget_exhausts():
    from poli import objective_factory

    problem = objective_factory.create(
        name="toy_continuous_problem",
        function_name="ackley_function_01",
        n_dimensions=3,
        evaluation_budget=10,
    )
    f = problem.black_box

    x0 = np.array([[0.0, 0.0, 0.0]] * 9)

    f(x0)

    with pytest.raises(BudgetExhaustedException):
        f(x0)



"""
This module implements LogP in _exactly_ the same was
as LaMBO does it [1]. We do this by importing the function
they use to compute `logp`.

References
----------
[1] “Accelerating Bayesian Optimization for Biological Sequence Design with Denoising Autoencoders.”
Stanton, Samuel, Wesley Maddox, Nate Gruver, Phillip Maffettone,
Emily Delaney, Peyton Greenside, and Andrew Gordon Wilson.
arXiv, July 12, 2022. http://arxiv.org/abs/2203.12742.
"""

import logging
import os
from pathlib import Path

import lambo
import numpy as np
from lambo import __file__ as project_root_file
from lambo.tasks.chem.logp import logP

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.util.chemistry.string_to_molecule import translate_selfies_to_smiles
from poli.core.util.files.download_files_from_github import (
    download_file_from_github_repository,
)

project_root = os.path.dirname(os.path.dirname(project_root_file))
LAMBO_IN_POLI_OBJECTIVES_PATH = Path.home() / ".poli_objectives" / "lambo"
LAMBO_IN_POLI_OBJECTIVES_PATH.mkdir(parents=True, exist_ok=True)

LAMBO_PACKAGE_ROOT = Path(lambo.__file__).parent.resolve()


def _download_assets_from_lambo():
    if os.environ.get("GITHUB_TOKEN_FOR_POLI") is None:
        logging.warning(
            "This black box objective function require downloading files "
            "from GitHub. Since the API rate limit is 60 requests per hour, "
            "we recommend creating a GitHub token and setting it as an "
            "environment variable called GITHUB_TOKEN_FOR_POLI. "
            "To create a GitHub token like this, follow the instructions here: "
            "https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/managing-your-personal-access-tokens#creating-a-fine-grained-personal-access-token"
        )

    # We should download files that are necessary, just like in
    # the other lambo objective function.

    # These files are lambo/tasks/chem/SA_Score/fpscores.pkl.gz
    fpscores_filepath = (
        LAMBO_PACKAGE_ROOT / "tasks" / "chem" / "SA_Score" / "fpscores.pkl.gz"
    )
    if not fpscores_filepath.exists():
        download_file_from_github_repository(
            "samuelstanton/lambo",
            "lambo/tasks/chem/SA_Score/fpscores.pkl.gz",
            str(fpscores_filepath),
            commit_sha="b8ea4e9",
            parent_folders_exist_ok=True,
            verbose=True,
        )


class PenalizedLogPIsolatedLogic(AbstractIsolatedFunction):
    def __init__(
        self,
        from_smiles: bool = True,
        penalized: bool = True,
    ):
        """
        TODO: document
        """
        self.from_smiles = from_smiles
        self.penalized = penalized
        _download_assets_from_lambo()

    def __call__(self, x: np.ndarray, context: dict = None):
        """
        Assuming that x is an array of strings (of shape [b,L]),
        we concatenate, translate to smiles if it's
        necessary, and then computes the penalized logP.

        If the inputs are SELFIES, it translates first to SMILES,
        and then computes the penalized logP. If the translation
        threw an error, we return NaN instead.
        """
        if x.dtype.kind not in ["U", "S"]:
            raise ValueError(
                f"We expect x to be an array of strings, but we got {x.dtype}"
            )

        molecule_strings = ["".join([x_ij for x_ij in x_i.flatten()]) for x_i in x]

        if not self.from_smiles:
            molecule_strings = translate_selfies_to_smiles(molecule_strings)

        logp_scores = []
        for molecule_string in molecule_strings:
            if molecule_string is None:
                logp_scores.append(np.nan)
            else:
                logp_scores.append(logP(molecule_string, penalized=self.penalized))

        return np.array(logp_scores).reshape(-1, 1)


if __name__ == "__main__":
    from poli.core.registry import register_isolated_function

    register_isolated_function(
        PenalizedLogPIsolatedLogic,
        name="penalized_logp_lambo__isolated",
        conda_environment_name="poli__lambo",
    )



"""RFP objective factory and black box function."""

__author__ = "Simon Bartels"

import logging
import os
from collections import namedtuple
from pathlib import Path

import hydra
import lambo
import numpy as np
import yaml
from lambo import __file__ as project_root_file
from lambo.utils import AMINO_ACIDS

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.registry import register_isolated_function
from poli.core.util.files.download_files_from_github import (
    download_file_from_github_repository,
)
from poli.objective_repository.foldx_rfp_lambo import CORRECT_SEQ, PROBLEM_SEQ

project_root = os.path.dirname(os.path.dirname(project_root_file))
LAMBO_IN_POLI_OBJECTIVES_PATH = Path.home() / ".poli_objectives" / "lambo"
LAMBO_IN_POLI_OBJECTIVES_PATH.mkdir(parents=True, exist_ok=True)

LAMBO_PACKAGE_ROOT = Path(lambo.__file__).parent.resolve()

LAMBO_FOLDX_ASSETS_PDBS = [
    "1uis_A",
    "1xqm_A",
    "1yzw_A",
    "1zgo_A",
    "1ztu_A",
    "2h5q_A",
    "2h5r_A",
    "2qlg_B",
    "2v4e_B",
    "2vad_A",
    "2vae_A",
    "2vvh_C",
    "3bxa_A",
    "3e5v_A",
    "3gb3_A",
    "3ip2_A",
    "3m22_A",
    "3ned_A",
    "3nez_A",
    "3nt3_C",
    "3nt9_A",
    "3p8u_C",
    "3pj5_A",
    "3rwa_A",
    "3s05_B",
    "3s7q_A",
    "3u0k_A",
    "3u0l_A",
    "3u8c_A",
    "4cqh_A",
    "4edo_A",
    "4eds_A",
    "4h3l_A",
    "4hq8_A",
    "4jf9_B",
    "4kge_A",
    "4oqw_A",
    "4p76_A",
    "4q7t_B",
    "4q7u_A",
    "5ajg_A",
    "5dtl_A",
    "5ez2_A",
    "5jva_A",
    "5lk4_A",
    "6aa7_A",
    "6dej_A",
    "6fzn_A",
    "6mgh_D",
    "6xwy_A",
]

tokenizer = {"_target_": "lambo.utils.ResidueTokenizer"}
Config = namedtuple("config", ["task", "tokenizer", "log_dir", "job_name", "timestamp"])


def get_config() -> Config:
    """
    Utility function with lambo specifc config to RFP task.
    """
    task = yaml.safe_load(
        (LAMBO_IN_POLI_OBJECTIVES_PATH / "proxy_rfp.yaml").read_text()
        # Path(
        #     str(Path(lambo.__file__).parent.resolve().parent.resolve())
        #     + os.path.sep
        #     + "hydra_config"
        #     + os.path.sep
        #     + "task"
        #     + os.path.sep
        #     + "proxy_rfp.yaml"
        # ).read_text()
    )
    config = Config(
        task,
        tokenizer,
        log_dir="data/experiments/test",
        job_name="null",
        timestamp="timestamp",
    )
    return config


def _download_assets_from_lambo():
    if os.environ.get("GITHUB_TOKEN_FOR_POLI") is None:
        logging.warning(
            "This black box objective function require downloading files "
            "from GitHub. Since the API rate limit is 60 requests per hour, "
            "we recommend creating a GitHub token and setting it as an "
            "environment variable called GITHUB_TOKEN_FOR_POLI. "
            "To create a GitHub token like this, follow the instructions here: "
            "https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/managing-your-personal-access-tokens#creating-a-fine-grained-personal-access-token"
        )

    # Downloading the config file into ~/.poli_objectives/lambo
    if not (LAMBO_IN_POLI_OBJECTIVES_PATH / "proxy_rfp.yaml").exists():
        download_file_from_github_repository(
            "samuelstanton/lambo",
            "hydra_config/task/proxy_rfp.yaml",
            str(Path.home() / ".poli_objectives" / "lambo" / "proxy_rfp.yaml"),
            commit_sha="431b052ad0e54a1ba4519272725310127c6377f1",
            parent_folders_exist_ok=True,
        )

    if not (
        LAMBO_PACKAGE_ROOT / "assets" / "fpbase" / "rfp_known_structures.csv"
    ).exists():
        download_file_from_github_repository(
            "samuelstanton/lambo",
            "lambo/assets/fpbase/rfp_known_structures.csv",
            str(LAMBO_PACKAGE_ROOT / "assets" / "fpbase" / "rfp_known_structures.csv"),
            commit_sha="431b052ad0e54a1ba4519272725310127c6377f1",
            parent_folders_exist_ok=True,
            verbose=True,
        )

    #     - proxy_rfp_seed_data.csv
    if not (
        LAMBO_PACKAGE_ROOT / "assets" / "fpbase" / "proxy_rfp_seed_data.csv"
    ).exists():
        download_file_from_github_repository(
            "samuelstanton/lambo",
            "lambo/assets/fpbase/proxy_rfp_seed_data.csv",
            str(LAMBO_PACKAGE_ROOT / "assets" / "fpbase" / "proxy_rfp_seed_data.csv"),
            commit_sha="431b052ad0e54a1ba4519272725310127c6377f1",
            parent_folders_exist_ok=True,
            verbose=True,
        )

    # - the sequences in the foldx folder.
    for folder_name in LAMBO_FOLDX_ASSETS_PDBS:
        if not (LAMBO_PACKAGE_ROOT / "assets" / "foldx" / folder_name).exists():
            download_file_from_github_repository(
                "samuelstanton/lambo",
                f"lambo/assets/foldx/{folder_name}",
                str(LAMBO_PACKAGE_ROOT / "assets" / "foldx" / folder_name),
                commit_sha="431b052ad0e54a1ba4519272725310127c6377f1",
                parent_folders_exist_ok=True,
                verbose=True,
            )


class RFPWrapperIsolatedLogic(AbstractIsolatedFunction):
    def __init__(
        self,
        seed: int = None,
    ):
        self.alphabet = AMINO_ACIDS
        self.problem_sequence = PROBLEM_SEQ
        self.correct_sequence = CORRECT_SEQ

        _download_assets_from_lambo()
        config = get_config()
        tokenizer = hydra.utils.instantiate(config.tokenizer)
        # NOTE: the task at this point is the original proxy rfp task
        bb_task = hydra.utils.instantiate(
            config.task, tokenizer=tokenizer, candidate_pool=[], seed=seed
        )
        base_candidates, base_targets, all_seqs, all_targets = bb_task.task_setup(
            config, project_root=project_root
        )

        # to ensure that algorithms will only request valid sequences, we provide the base sequences first
        permutation = np.arange(all_seqs.shape[0])
        base_candidate_idx = np.array([0, 3, 4, 16, 37, 39])
        permutation[base_candidate_idx] = np.arange(base_candidate_idx.shape[0])
        permutation[: base_candidate_idx.shape[0]] = base_candidate_idx
        all_seqs = all_seqs[permutation, ...]
        all_targets = all_targets[permutation, ...]
        assert np.all(all_targets[: base_targets.shape[0], ...] == base_targets)

        # all_seqs should be of shape [b, L], adding empty padding
        max_seq_length = max([len(seq) for seq in all_seqs])
        all_seqs = np.array(
            [list(seq) + [""] * (max_seq_length - len(seq)) for seq in all_seqs]
        )
        if self.problem_sequence in all_seqs:
            # substitute erroneous sequence "X in position 159" with correct PDB fasta sequence
            all_seqs[np.where(all_seqs == self.problem_sequence)] = (
                self.correct_sequence
            )

        self.task = bb_task
        self.base_candidates = base_candidates
        self.sequences_aligned = False
        self.inverse_alphabet = {i + 1: AMINO_ACIDS[i] for i in range(len(AMINO_ACIDS))}
        self.inverse_alphabet[0] = "-"

        self.x0 = all_seqs
        self.y0 = all_targets

    def __call__(self, x, context=None):
        best_b_cand = None
        min_hd = np.infty  # Hamming distance of best_b_cand to x

        # TODO: this assumes a batch_size of 1. Is that clear in the docs?
        seq = "".join(x[0])  # take out the string from the np array
        for b_cand in self.base_candidates:
            b_seq = b_cand.mutant_residue_seq
            if b_seq is None:
                raise ValueError(
                    "Base candidates (from lambo's FoldedCandidates) is empty."
                    "\nThis usually happens when the internal foldx process fails."
                    "Verify your foldx installation and try again."
                )
            if len(b_seq) != len(seq):
                continue
            hd = np.sum([seq[i] != b_seq[i] for i in range(len(seq))])
            if hd < min_hd:
                min_hd = hd
                best_b_cand = b_cand
        if best_b_cand is None:
            for b_cand in self.base_candidates:
                logging.fatal(
                    "%i is a valid candidate length." % len(b_cand.mutant_residue_seq)
                )
            raise RuntimeError("Cannot evaluate candidate of length %i." % len(seq))
        return self.task.score(
            self.task.make_new_candidates(np.array([best_b_cand]), np.array([seq]))
        )


if __name__ == "__main__":
    # inner_function = RFPWrapperIsolatedLogic()

    register_isolated_function(
        RFPWrapperIsolatedLogic,
        name="foldx_rfp_lambo__isolated",
        conda_environment_name="poli__lambo",
    )



"""poli, a library for discrete black-box objective functions."""

__version__ = "1.1.1"
from .core.util.isolation.instancing import instance_function_as_isolated_process

# from .core import get_problems
from .objective_factory import create
from .objective_repository import get_problems

__all__ = ["create", "get_problems", "instance_function_as_isolated_process"]



from poli.core.black_box_information import BlackBoxInformation

SMB_ALPHABET = [
    "X",
    "S",
    "-",
    "?",
    "Q",
    "E",
    "<",
    ">",
    "[",
    "]",
    "o",
]

smb_info = BlackBoxInformation(
    name="super_mario_bros",
    max_sequence_length=14 * 14,
    aligned=True,
    fixed_length=True,
    deterministic=False,
    alphabet=SMB_ALPHABET,
    log_transform_recommended=True,
    discrete=True,
    padding_token=None,
)



"""This module tests the instancing of toy continuous problems."""

import numpy as np
import pytest

from poli.objective_repository.toy_continuous_problem.toy_continuous_problem import (
    POSSIBLE_FUNCTIONS,
    SIX_DIMENSIONAL_PROBLEMS,
    TWO_DIMENSIONAL_PROBLEMS,
)


@pytest.mark.parametrize("function_name", POSSIBLE_FUNCTIONS)
def test_create_toy_objective_function(function_name):
    """Tests the instancing the given objective function."""
    from poli import objective_factory

    if function_name in SIX_DIMENSIONAL_PROBLEMS:
        n_dimensions = 6
    elif function_name in TWO_DIMENSIONAL_PROBLEMS:
        n_dimensions = 2
    else:
        n_dimensions = 10

    problem = objective_factory.create(
        name="toy_continuous_problem",
        function_name=function_name,
        n_dimensions=n_dimensions,
    )
    f = problem.black_box
    x0 = problem.x0

    f(x0)


def test_create_ackley_function_01_on_more_dimensions():
    """Tests the instancing of the Ackley function in 2D."""
    from poli import objective_factory

    problem = objective_factory.create(
        name="toy_continuous_problem",
        function_name="ackley_function_01",
        n_dimensions=10,
    )
    f_ackley = problem.black_box
    f_ackley(np.array([[0.0] * 10]))



"""
Implements a synthetic-accessibility objective using the TDC oracles [1].

References
----------
[1] Artificial intelligence foundation for therapeutic science.
    Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022). https://doi.org/10.1038/s41589-022-01131-2
"""

from poli.core.chemistry.tdc_isolated_function import TDCIsolatedFunction


class SAIsolatedLogic(TDCIsolatedFunction):
    """Synthetic-accessibility black box implementation using the TDC oracles [1].

    Parameters
    ----------
    from_smiles : bool, optional
        Flag indicating whether to use SMILES strings as input, by default True.
    """

    def __init__(
        self,
        from_smiles: bool = True,
    ):
        """
        Initialize the SABlackBox object.

        Parameters
        ----------
        from_smiles : bool, optional
            Flag indicating whether to use SMILES strings as input, by default True.
        """
        oracle_name = "SA"
        super().__init__(
            oracle_name=oracle_name,
            from_smiles=from_smiles,
        )


if __name__ == "__main__":
    from poli.core.registry import register_isolated_function

    register_isolated_function(
        SAIsolatedLogic,
        name="sa_tdc__isolated",
        conda_environment_name="poli__tdc",
        force=True,
    )



"""Registers the white noise problem factory and black box.

This is a registration script for the white_noise problem,
whose black box objective function returns standard Gaussian
noise.

The problem is registered as 'white_noise', and it uses
a conda environment called 'poli__base' (see the
environment.yml file in this folder).
"""

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.seeding import seed_python_numpy_and_torch


class WhiteNoiseBlackBox(AbstractBlackBox):
    """
    A toy black box function that generates standard Gaussian noise.

    Parameters
    ----------
    batch_size : int, optional
        The batch size for vectorized evaluation.
    parallelize : bool, optional
        Whether to parallelize the evaluation.
    num_workers : int, optional
        The number of workers for parallel evaluation.
    evaluation_budget : int, optional
        The maximum number of evaluations.

    Methods
    -------
    _black_box(x, context=None)
        Returns standard Gaussian noise.

    """

    def __init__(
        self,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
    ):
        """
        Initializes a WhiteNoiseBlackBox.

        Parameters
        ----------
        info : ProblemSetupInformation
            The problem setup information.
        batch_size : int, optional
            The batch size for vectorized evaluation, by default None (i.e.
            all of the input).
        parallelize : bool, optional
            Whether to parallelize the evaluation, by default False.
        num_workers : int, optional
            The number of workers for parallel evaluation, by default None (which
            corresponds to half the CPUs available, rounded downwards).
        evaluation_budget : int, optional
            The maximum number of evaluations, by default float("inf").
        """
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

    def _black_box(self, x: np.ndarray, context: dict = None) -> np.ndarray:
        """Returns standard Gaussian noise.

        Parameters
        ----------
        x : np.ndarray
            The input.
        context : dict, optional
            The context, by default None.

        Returns
        -------
        y : np.ndarray
            Standard Gaussian noise.
        """
        return np.random.randn(x.shape[0], 1)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="white_noise",
            max_sequence_length=np.inf,
            aligned=False,
            fixed_length=False,
            deterministic=False,
            alphabet=[str(i) for i in range(10)],
            log_transform_recommended=False,
            discrete=True,
            padding_token="",
        )


class WhiteNoiseProblemFactory(AbstractProblemFactory):
    def create(
        self,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Create a white noise problem with the specified parameters.

        Parameters:
        ----------
        seed : int, optional
            The seed value for random number generation. If not provided, no seeding will be performed.
        batch_size : int, optional
            The number of samples to evaluate each time. If not provided, the default batch size will be used.
        parallelize : bool, optional
            Whether to parallelize the evaluation of samples. Defaults to False.
        num_workers : int, optional
            The number of worker processes to use for parallel evaluation. If not provided,
            half the available CPUs will be used (rounded down).
        evaluation_budget : int, optional
            The maximum number of evaluations allowed. Defaults to infinity.

        Returns:
        -------
        problem: WhiteNoiseProblem
            A white noise problem with the specified parameters.
        """
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        f = WhiteNoiseBlackBox(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        x0 = np.array([["1", "2", "3"]])

        white_noise_problem = Problem(black_box=f, x0=x0)

        return white_noise_problem



from pathlib import Path

import numpy as np
import pytest

from poli import objective_factory

THIS_DIR = Path(__file__).parent.resolve()


@pytest.mark.poli__rasp
def test_rasp_on_3ned_against_notebooks_results():
    problem = objective_factory.create(
        name="rasp",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
    )
    f, x0 = problem.black_box, problem.x0

    wildtype_sequence = "".join(x0[0])
    three_mutations = [
        "A" + wildtype_sequence[1:],
        "R" + wildtype_sequence[1:],
        "N" + wildtype_sequence[1:],
    ]

    x = np.array([list(mutation) for mutation in three_mutations])
    y = f(x)

    # Asserting that the results are the same as in the
    # notebook:
    # E1A: 0.03654138690753095
    # E1R: -0.07091977827871465
    # E1N: -0.2835593180137258

    assert np.isclose(y[0], -0.03654138690753095, atol=1e-4)
    assert np.isclose(y[1], 0.07091977827871465, atol=1e-4)
    assert np.isclose(y[2], 0.2835593180137258, atol=1e-4)


@pytest.mark.poli__rasp
def test_rasp_on_3ned_evaluating_twice_vs_once():
    problem = objective_factory.create(
        name="rasp",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
    )
    f, x0 = problem.black_box, problem.x0

    wildtype_sequence = "".join(x0[0])
    three_mutations = [
        "A" + wildtype_sequence[1:],
        "R" + wildtype_sequence[1:],
        "N" + wildtype_sequence[1:],
    ]

    two_mutations = [
        "A" + wildtype_sequence[1:],
        "R" + wildtype_sequence[1:],
    ]
    one_mutation = [
        "N" + wildtype_sequence[1:],
    ]

    x_all = np.array([list(mutation) for mutation in three_mutations])
    y = f(x_all)

    y_two = f(np.array([list(mutation) for mutation in two_mutations]))
    y_one = f(np.array([list(mutation) for mutation in one_mutation]))

    assert y[0] == y_two[0]
    assert y[1] == y_two[1]
    assert y[2] == y_one[0]
    # Asserting that the results are the same as in the
    # notebook:
    # E1A: 0.03654138690753095
    # E1R: -0.07091977827871465
    # E1N: -0.2835593180137258

    assert np.isclose(y[0], -0.03654138690753095, atol=1e-4)
    assert np.isclose(y[1], 0.07091977827871465, atol=1e-4)
    assert np.isclose(y[2], 0.2835593180137258, atol=1e-4)


@pytest.mark.poli__rasp
def test_rasp_on_3ned_against_notebooks_results_isolated():
    """
    We test forceful registration of the RaSP problem.
    """
    problem = objective_factory.create(
        name="rasp",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        force_isolation=True,
    )
    f, x0 = problem.black_box, problem.x0

    wildtype_sequence = "".join(x0[0])
    three_mutations = [
        "A" + wildtype_sequence[1:],
        "R" + wildtype_sequence[1:],
        "N" + wildtype_sequence[1:],
    ]

    x = np.array([list(mutation) for mutation in three_mutations])
    y = f(x)

    # Asserting that the results are the same as in the
    # notebook:
    # E1A: 0.03654138690753095
    # E1R: -0.07091977827871465
    # E1N: -0.2835593180137258

    # Notice how we are clipping the actual values,
    # this is because we would need double precision
    # to test against the exact values described above.
    # TODO: Should we do double precision by default
    # inside the RaSP problem?
    assert np.isclose(y[0], -0.0365, atol=1e-4)
    assert np.isclose(y[1], 0.07091, atol=1e-4)
    assert np.isclose(y[2], 0.283559, atol=1e-4)


@pytest.mark.poli__rasp
def test_rasp_using_additive_flag_on_two_mutations():
    problem = objective_factory.create(
        name="rasp",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        additive=True,
    )
    f, x0 = problem.black_box, problem.x0

    wildtype_sequence = "".join(x0[0])
    one_mutant_with_two_mutations = [
        "AR" + wildtype_sequence[2:],
    ]
    two_mutations = [
        "A" + wildtype_sequence[1:],
        wildtype_sequence[:1] + "R" + wildtype_sequence[2:],
    ]

    x = np.array([list(mutation) for mutation in one_mutant_with_two_mutations])
    y = f(x)

    x1 = np.array([list(mutation) for mutation in two_mutations])
    y1 = f(x1)

    assert np.isclose(y, y1.sum(), atol=1e-4)


@pytest.mark.poli__rasp
def test_rasp_penalization_works():
    problem = objective_factory.create(
        name="rasp",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        additive=True,
        penalize_unfeasible_with=-100.0,
    )
    f, _ = problem.black_box, problem.x0

    # This is an unfeasible mutation, since joining
    # all the strings would result in a sequence
    # that is _not_ the same length as the wildtype.
    problematic_x = np.array([["A"] + [""] * (f.info.max_sequence_length - 1)])
    assert f(problematic_x) == -100.0


@pytest.mark.poli__rasp
def test_rasp_penalization_works_on_multiple_inputs():
    problem = objective_factory.create(
        name="rasp",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        additive=True,
        penalize_unfeasible_with=-100.0,
    )
    f, _ = problem.black_box, problem.x0

    # This is an unfeasible mutation, since joining
    # all the strings would result in a sequence
    # that is _not_ the same length as the wildtype.
    problematic_x = np.array([["A"] + [""] * (f.info.max_sequence_length - 1)])
    combination = np.vstack([problem.x0, problematic_x])
    y = f(combination)
    assert y[-1] == -100.0


@pytest.mark.poli__rasp
def test_rasp_fails_on_invalid_amino_acids():
    problem = objective_factory.create(
        name="rasp",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
    )
    f, x0 = problem.black_box, problem.x0

    # This is an unfeasible mutation, since joining
    # all the strings would result in a sequence
    # that is _not_ the same length as the wildtype.
    problematic_x = x0.copy()
    problematic_x[0][0] = "B"
    with pytest.raises(ValueError):
        f(problematic_x)



import numpy as np

from poli.core.black_box_information import BlackBoxInformation
from poli.core.util.proteins.defaults import AMINO_ACIDS

rosetta_energy_information = BlackBoxInformation(
    name="rosetta_energy",
    max_sequence_length=np.inf,
    alphabet=AMINO_ACIDS,
    aligned=True,
    fixed_length=True,
    discrete=True,
    deterministic=False,
    padding_token="",
)



from poli.repository import ToyContinuousBlackBox


def test_dynamic_info_on_toy_continuous_black_box():
    f = ToyContinuousBlackBox(
        function_name="ackley_function_01",
        n_dimensions=4,
    )

    assert f.info.max_sequence_length == 4



"""
Registers the stability and SASA FoldX black box
and objective factory.

FoldX [1] is a simulator that allows for computing the difference
in free energy between a wildtype protein and a mutated protein.
We pair this with biopython [2] to compute the SASA of the mutated
protein.

[1] Schymkowitz, J., Borg, J., Stricher, F., Nys, R., Rousseau, F.,
    & Serrano, L. (2005). The FoldX web server: an online force field.
    Nucleic acids research, 33(suppl_2), W382-W388.
[2] Cock PA, Antao T, Chang JT, Chapman BA, Cox CJ, Dalke A, Friedberg I,
    Hamelryck T, Kauff F, Wilczynski B and de Hoon MJL (2009) Biopython:
    freely available Python tools for computational molecular biology and
    bioinformatics. Bioinformatics, 25, 1422-1423
"""

from pathlib import Path
from typing import List, Union

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.exceptions import FoldXNotFoundException
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import get_inner_function
from poli.core.util.proteins.defaults import AMINO_ACIDS
from poli.core.util.seeding import seed_python_numpy_and_torch


class FoldXStabilityAndSASABlackBox(AbstractBlackBox):
    """
    A black box implementation for computing the solvent accessible surface area (SASA) score using FoldX.

    Parameters
    -----------
    wildtype_pdb_path : Union[Path, List[Path]]
        The path(s) to the wildtype PDB file(s).
    experiment_id : str, optional
        The ID of the experiment. Default is None.
    tmp_folder : Path, optional
        The path to the temporary folder. Default is None.
    eager_repair : bool, optional
        Whether to perform eager repair. Default is False.
    verbose : bool, optional
        Whether to print the output from FoldX. Default is False.
    batch_size : int, optional
        The batch size for parallel processing. Default is None.
    parallelize : bool, optional
        Whether to parallelize the computation. Default is False.
    num_workers : int, optional
        The number of workers for parallel processing. Default is None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.

    Notes
    -----
    We expect the user to have FoldX v5.0 installed and compiled.
    More specifically, we expect a binary file called foldx to be
    in the path ~/foldx/foldx.
    """

    def __init__(
        self,
        wildtype_pdb_path: Union[Path, List[Path]],
        experiment_id: str = None,
        tmp_folder: Path = None,
        eager_repair: bool = False,
        verbose: bool = False,
        batch_size: int = 1,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.wildtype_pdb_path = wildtype_pdb_path
        self.experiment_id = experiment_id
        self.tmp_folder = tmp_folder
        self.eager_repair = eager_repair
        self.verbose = verbose
        self.force_isolation = force_isolation
        if not (Path.home() / "foldx" / "foldx").exists():
            raise FoldXNotFoundException(
                "FoldX wasn't found in ~/foldx/foldx. Please install it."
            )
        inner_function = get_inner_function(
            isolated_function_name="foldx_stability_and_sasa__isolated",
            class_name="FoldXStabilitityAndSASAIsolatedLogic",
            module_to_import="poli.objective_repository.foldx_stability_and_sasa.isolated_function",
            wildtype_pdb_path=wildtype_pdb_path,
            experiment_id=experiment_id,
            tmp_folder=tmp_folder,
            eager_repair=eager_repair,
            verbose=verbose,
        )
        self.wildtype_amino_acids = inner_function.wildtype_amino_acids

    def _black_box(self, x: np.ndarray, context: None) -> np.ndarray:
        """
        Runs the given input x and pdb files provided
        in the context through FoldX and returns the
        total energy score.

        Since the goal is MINIMIZING the energy,
        we return the negative of the total energy.

        Parameters
        -----------
        x : np.ndarray
            The input array representing the mutations.
        context : None
            The context for the black box computation.

        Returns
        --------
        y: np.ndarray
            The computed stability and SASA score(s) as a numpy array.
        """
        inner_function = get_inner_function(
            isolated_function_name="foldx_stability_and_sasa__isolated",
            class_name="FoldXStabilitityAndSASAIsolatedLogic",
            module_to_import="poli.objective_repository.foldx_stability_and_sasa.isolated_function",
            quiet=True,
            wildtype_pdb_path=self.wildtype_pdb_path,
            experiment_id=self.experiment_id,
            tmp_folder=self.tmp_folder,
            eager_repair=self.eager_repair,
            verbose=self.verbose,
        )

        return inner_function(x, context)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="foldx_stability_and_sasa",
            max_sequence_length=np.inf,
            aligned=False,
            fixed_length=False,
            deterministic=True,
            alphabet=AMINO_ACIDS,
            log_transform_recommended=False,
            discrete=True,
            fidelity=None,
            padding_token="",
        )


class FoldXStabilityAndSASAProblemFactory(AbstractProblemFactory):
    """
    Factory class for creating FoldX SASA (Solvent Accessible Surface Area) problems.

    Methods
    -------
    create:
        Creates a problem instance with the specified parameters.
    """

    def create(
        self,
        wildtype_pdb_path: Union[Path, List[Path]],
        experiment_id: str = None,
        tmp_folder: Path = None,
        eager_repair: bool = False,
        verbose: bool = False,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Create a FoldXSASABlackBox object and compute the initial values of wildtypes.

        Parameters
        ----------
        wildtype_pdb_path : Union[Path, List[Path]]
            Path or list of paths to the wildtype PDB files.
        experiment_id : str, optional
            Identifier for the experiment.
        tmp_folder : Path, optional
            Path to the temporary folder for intermediate files.
        eager_repair : bool, optional
            Flag indicating whether to perform eager repair.
        verbose : bool, optional
            Flag indicating whether to print the output from FoldX.
        seed : int, optional
            Seed for random number generators. If None is passed,
            the seeding doesn't take place.
        batch_size : int, optional
            Number of samples per batch for parallel computation.
        parallelize : bool, optional
            Flag indicating whether to parallelize the computation.
        num_workers : int, optional
            Number of worker processes for parallel computation.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.

        Returns
        -------
        problem : Problem
            A problem instance containing a FoldXStabilityAndSASABlackBox
            function, and initial wildtypes x0.

        Raises
        ------
        ValueError
            If wildtype_pdb_path is missing or has an invalid type.
        """
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        if wildtype_pdb_path is None:
            raise ValueError(
                "Missing required argument wildtype_pdb_path. "
                "Did you forget to pass it to create()?"
            )

        if isinstance(wildtype_pdb_path, str):
            wildtype_pdb_path = [Path(wildtype_pdb_path.strip())]
        elif isinstance(wildtype_pdb_path, Path):
            wildtype_pdb_path = [wildtype_pdb_path]
        elif isinstance(wildtype_pdb_path, list):
            if isinstance(wildtype_pdb_path[0], str):
                wildtype_pdb_path = [Path(x.strip()) for x in wildtype_pdb_path]
            elif isinstance(wildtype_pdb_path[0], Path):
                pass
        else:
            raise ValueError(
                f"wildtype_pdb_path must be a string or a Path. Received {type(wildtype_pdb_path)}"
            )
        # By this point, we know that wildtype_pdb_path is a
        # list of Path objects.

        # We use the default alphabet if None was provided.
        # See ENCODING in foldx_utils.py
        f = FoldXStabilityAndSASABlackBox(
            wildtype_pdb_path=wildtype_pdb_path,
            experiment_id=experiment_id,
            tmp_folder=tmp_folder,
            eager_repair=eager_repair,
            verbose=verbose,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )

        # We need to compute the initial values of all wildtypes
        # in wildtype_pdb_path. For this, we need to specify x0,
        # a vector of wildtype sequences. These are padded to
        # match the maximum length with empty strings.
        wildtype_amino_acids_ = f.wildtype_amino_acids

        # # We may need to parallelize. If so, the inner_function object
        # # can't be pickled. If that's the case, we need to delete it.
        # # Each call to the black box will create its own connection
        # # to an isolated function.
        # del f.inner_function

        longest_wildtype_length = max([len(x) for x in wildtype_amino_acids_])

        wildtype_amino_acids = [
            amino_acids + [""] * (longest_wildtype_length - len(amino_acids))
            for amino_acids in wildtype_amino_acids_
        ]

        x0 = np.array(wildtype_amino_acids).reshape(
            len(wildtype_pdb_path), longest_wildtype_length
        )

        problem = Problem(f, x0)
        return problem



"""An interface to the original RaSP codebase.

This module takes and adapts RaSP's original implementation
(which can be found at [1]), and writes an interface that
handles the preprocessing and inference steps.

RaSP, like foldx [2], predicts the effect of mutations on protein
stability. However, it does so using a deep learning model
instead of actual simulations. The drawback being that
only additive mutations are supported (indeed, we currently
only support one-mutation-from-wildtype).

To perform inference, RaSP requires downloading the following:
- the `reduce` executable, which is used to clean the PDB files.
- the cavity model, which is used to predict the cavity features.
- the downstream models, which are used to predict the ddG values.

This means that you will need internet access to use this interface
(and the "rasp" objective function) for the first time. These models
and tools will be cached in the ~/.poli_objectives/rasp directory,
so you will only need internet access once.

RaSP's source code was provided with an Apache 2.0 license. Modifications
from the source have been duly noted. Most of the source code can be found
at ./inner_rasp.

References
----------
[1] “Rapid Protein Stability Prediction Using Deep Learning Representations.”
    Blaabjerg, Lasse M, Maher M Kassem, Lydia L Good, Nicolas Jonsson, Matteo Cagiada,
    Kristoffer E Johansson, Wouter Boomsma, Amelie Stein, and Kresten Lindorff-Larsen.
    Edited by José D Faraldo-Gómez, Detlef Weigel, Nir Ben-Tal, and Julian Echave.
    eLife 12 (May 2023): e82593. https://doi.org/10.7554/eLife.82593.
[2] The FoldX web server: an online force field.
    Schymkowitz, J., Borg, J., Stricher, F., Nys, R.,
    Rousseau, F., & Serrano, L. (2005).  Nucleic acids research,
    33(suppl_2), W382-W388.
"""

import logging
import os
import stat
import subprocess
import traceback
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
from Bio.PDB.Polypeptide import index_to_one, one_to_index
from pdbtools.pdb_delhetatm import run as pdb_delhetatm_run
from pdbtools.pdb_delres import run as pdb_delres_run
from pdbtools.pdb_fixinsert import run as pdb_fixinsert_run
from pdbtools.pdb_selchain import run as pdb_selchain_run
from pdbtools.pdb_tidy import run as pdb_tidy_run

from poli.core.util.files.download_files_from_github import (
    download_file_from_github_repository,
)
from poli.core.util.files.integrity import compute_md5_from_filepath
from poli.core.util.proteins.mutations import edits_between_strings
from poli.core.util.proteins.rasp.inner_rasp.cavity_model import (
    ResidueEnvironmentsDataset,
)
from poli.core.util.proteins.rasp.inner_rasp.helpers import ds_pred
from poli.core.util.proteins.rasp.inner_rasp.pdb_parser_scripts.clean_pdb import (
    clean_pdb,
)
from poli.core.util.proteins.rasp.inner_rasp.pdb_parser_scripts.extract_environments import (
    extract_environments,
)

THIS_DIR = Path(__file__).parent.resolve()
HOME_DIR = THIS_DIR.home()
RASP_DIR = HOME_DIR / ".poli_objectives" / "rasp"
RASP_DIR.mkdir(parents=True, exist_ok=True)


class RaspInterface:
    def __init__(
        self,
        working_dir: Path,
        verbose: bool = False,
        verify_integrity_of_download: bool = True,
    ) -> None:
        self.working_dir = working_dir
        self.verbose = verbose
        self.verify_integrity_of_download = verify_integrity_of_download

        # Making the appropriate folders:
        (self.working_dir / "raw").mkdir(parents=True, exist_ok=True)
        (self.working_dir / "cleaned").mkdir(parents=True, exist_ok=True)
        (self.working_dir / "parsed").mkdir(parents=True, exist_ok=True)
        (self.working_dir / "output").mkdir(parents=True, exist_ok=True)

        # Downloading reduce and compile it.
        # TODO: should we be doing this eagerly? Or should
        # we wait to download and install it until we need it?
        self.reduce_executable_path = None
        self.get_and_compile_reduce()

        # At this point, we should have the reduce executable
        # at self.reduce_executable_path.

        # Downloading the cavity and downstream models.
        # TODO: should we be doing this eagerly? Or should
        # we wait to download and install it until we need it?
        self.download_cavity_and_downstream_models(
            verbose=verbose, verify_integrity_of_download=verify_integrity_of_download
        )
        self.alphabet = [
            "A",
            "C",
            "D",
            "E",
            "F",
            "G",
            "H",
            "I",
            "K",
            "L",
            "M",
            "N",
            "P",
            "Q",
            "R",
            "S",
            "T",
            "V",
            "W",
            "Y",
        ]

    def get_and_compile_reduce(self):
        """
        This function downloads reduce and compiles it
        if we can't find it inside ~/.poli_objectives/rasp.
        """
        HOME_DIR = self.working_dir.home()
        RASP_DIR = HOME_DIR / ".poli_objectives" / "rasp"
        RASP_DIR.mkdir(parents=True, exist_ok=True)

        REDUCE_DIR = RASP_DIR / "reduce"
        EXECUTABLE_PATH = REDUCE_DIR / "reduce_src" / "reduce"
        if not (EXECUTABLE_PATH).exists():
            # Clone it using git.
            # TODO: Is there a way of downloading the contents
            # of the repo without cloning it?
            try:
                subprocess.run(
                    "git clone https://github.com/rlabduke/reduce.git",
                    cwd=RASP_DIR,
                    shell=True,
                    stdout=subprocess.DEVNULL,
                    check=True,
                )

                # Pinning to commit hash bd23a0bf...
                subprocess.run(
                    "git checkout bd23a0bf627ae9b08842102a5c2e9404b4a81924",
                    cwd=REDUCE_DIR,
                    shell=True,
                    stdout=subprocess.DEVNULL,
                    check=True,
                )
            except subprocess.CalledProcessError:
                # Printing the traceback
                traceback.print_exc()
                raise RuntimeError(
                    "Could not clone the reduce repository. "
                    "Please check your internet connection."
                )

            # compile it.
            try:
                subprocess.run(
                    "make",
                    cwd=REDUCE_DIR,
                    stdout=subprocess.DEVNULL,
                    check=True,
                )
            except subprocess.CalledProcessError:
                # TODO: should we be purging it ourselves?
                traceback.print_exc()
                raise RuntimeError(
                    "Something went wrong while compiling reduce. "
                    "Purge the folder ~/.poli_objectives/rasp/reduce "
                    " and try again."
                )

            # Change its permissions.
            os.chmod(EXECUTABLE_PATH, stat.S_IEXEC)

        self.reduce_executable_path = EXECUTABLE_PATH

    def predict(
        self,
        cavity_model_net,
        ds_model_net,
        df_structure,
        dataset_key,
        NUM_ENSEMBLE,
        DEVICE,
    ) -> pd.DataFrame:
        df_ml = ds_pred(
            cavity_model_net,
            ds_model_net,
            df_structure,
            dataset_key,
            NUM_ENSEMBLE,
            DEVICE,
        )

        df_total = df_structure.merge(
            df_ml, on=["pdbid", "chainid", "variant"], how="outer"
        )
        df_total = df_total.drop("resenv", axis=1)

        # Removed by MGD: we don't need the PRISM files, we only
        # need the dataframe.
        #
        # Parse output into separate files by pdb, print to PRISM format
        # for pdbid in df_total["pdbid"].unique():
        #     df_pdb = df_total[df_total["pdbid"] == pdbid]
        #     for chainid in df_pdb["chainid"].unique():
        #         pred_outdir = self.working_dir / "output" / f"{dataset_key}"
        #         pred_outdir.mkdir(parents=True, exist_ok=True)
        #         pred_outfile = pred_outdir / f"cavity_pred_{pdbid}_{chainid}.csv"
        #         print(
        #             f"Parsing predictions from pdb: {pdbid}{chainid} into {pred_outfile}"
        #         )
        #         df_chain = df_pdb[df_pdb["chainid"] == chainid]
        #         df_chain = df_chain.assign(pos=df_chain["variant"].str[1:-1])
        #         df_chain["pos"] = pd.to_numeric(df_chain["pos"])
        #         first_res_no = min(df_chain["pos"])
        #         df_chain = df_chain.assign(wt_AA=df_chain["variant"].str[0])
        #         df_chain = df_chain.assign(mt_AA=df_chain["variant"].str[-1])
        #         seq = get_seq_from_variant(df_chain)
        #         df_chain.to_csv(pred_outfile, index=False)
        #         prism_outfile = pred_outdir / f"prism_cavity_{pdbid}_{chainid}.txt"

        #         # if (AF_ID !=''):
        #         #   prism_outfile = f"/content/output/{dataset_key}/prism_cavity_{AF_ID}_{chainid}.txt"
        #         # elif (PDB_ID !=''):
        #         #   prism_outfile = f"/content/output/{dataset_key}/prism_cavity_{PDB_ID}_{chainid}.txt"
        #         # elif PDB_custom:
        #         #   prism_outfile = f"/content/output/{dataset_key}/prism_cavity_XXXX_{chainid}.txt"
        #         # cavity_to_prism(df_chain, pdbid, chainid, seq, prism_outfile)

        return df_total

    def download_cavity_and_downstream_models(
        self, verbose: bool = False, verify_integrity_of_download: bool = True
    ):
        """
        This function downloads the cavity and downstream models
        at the ~/.poli_objectives/rasp directory.

        If strict is True, then we will raise an error if the
        models we download don't match the expected MD5 checksums.
        Otherwise, we will just log a warning.
        """
        if os.environ.get("GITHUB_TOKEN_FOR_POLI") is None:
            logging.warning(
                "This black box objective function require downloading files "
                "from GitHub. Since the API rate limit is 60 requests per hour, "
                "we recommend creating a GitHub token and setting it as an "
                "environment variable called GITHUB_TOKEN_FOR_POLI. "
                "To create a GitHub token like this, follow the instructions here: "
                "https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/managing-your-personal-access-tokens#creating-a-fine-grained-personal-access-token"
            )

        cavity_model_path = RASP_DIR / "cavity_model_15.pt"
        ds_models_paths = [
            RASP_DIR / "ds_models" / f"ds_model_{i}" / "model.pt" for i in range(10)
        ]

        repository_name = "KULL-Centre/papers"
        output_folder_in_repository = "2022/ML-ddG-Blaabjerg-et-al/output"
        commit_sha = "3ccebe87e017b6bd737f88e1943557d128c85616"

        # We will compare the md5 checksums of the models
        # we download against these pre-computed md5s.
        precomputed_model_md5_checksums = {
            cavity_model_path: "7f039dab359da67f5870e4c8ba4204ee",
            "ds_model_0": "88d8d836f4070245d23532ec46ff886b",
            "ds_model_1": "4ceca5dda221251f1333cc67717339f2",
            "ds_model_2": "9126cb60c340d2c165e05e7a8e5b3ffa",
            "ds_model_3": "d3cc21c451b1e5dbe797937fd6c34d3e",
            "ds_model_4": "a0d23110519a3763205a795fb713b5bf",
            "ds_model_5": "6c0bc930643aa51ebeba44ddea3b5865",
            "ds_model_6": "1ad01d6e52160f9f4083775f247d6127",
            "ds_model_7": "2470800d25b8c3c088c1eb17e6c2e440",
            "ds_model_8": "313b25729790e7698ce9f35b0b632177",
            "ds_model_9": "9f02a8de64a4a354eb4734d54ee9ba23",
        }

        if verbose:
            print("poli 🧪: Downloading the cavity and downstream models")
            print(f"Repository: {repository_name}")
            print(f"Commit: {commit_sha}")

        # Downloading the cavity model.
        if not cavity_model_path.exists():
            if verbose:
                print("poli 🧪: Downloading the cavity model")

            download_file_from_github_repository(
                repository_name=repository_name,
                file_path_in_repository=f"{output_folder_in_repository}/cavity_models/cavity_model_15.pt",
                download_path_for_file=cavity_model_path,
                commit_sha=commit_sha,
                exist_ok=True,
                verbose=verbose,
            )

            # Verifying the integrity of the file we just downloaded.
            downloaded_md5_checksum = compute_md5_from_filepath(cavity_model_path)
            if (
                downloaded_md5_checksum
                != precomputed_model_md5_checksums[cavity_model_path]
            ):
                if verify_integrity_of_download:
                    raise RuntimeError(
                        "The downloaded cavity model does not match the expected md5 checksum. "
                        "This could be due to a corrupted download, or a malicious attack. "
                        "Delete the files at ~/.poli_objectives/rasp and try again."
                    )
                else:
                    logging.warning(
                        "The downloaded cavity model does not match the expected md5 checksum. "
                        "This could be due to a corrupted download, or a malicious attack. "
                    )
        elif verbose:
            print("poli 🧪: Cavity model already exists. Skipping.")

        # Downloading the downstream models
        for path_ in ds_models_paths:
            if not path_.exists():
                if verbose:
                    print(f"poli 🧪: Downloading the downstream model {path_.parent}")

                local_path_in_directory = path_.relative_to(RASP_DIR)
                download_file_from_github_repository(
                    repository_name=repository_name,
                    file_path_in_repository=f"{output_folder_in_repository}/{local_path_in_directory}",
                    download_path_for_file=path_,
                    commit_sha=commit_sha,
                    exist_ok=False,
                )

                # Verifying the integrity of the file we just downloaded.
                downloaded_md5_checksum = compute_md5_from_filepath(path_)
                if (
                    downloaded_md5_checksum
                    != precomputed_model_md5_checksums[path_.parent.name]
                ):
                    if verify_integrity_of_download:
                        raise RuntimeError(
                            f"The downloaded downstream model {path_.parent.name} does not match the expected md5 checksum. "
                            "This could be due to a corrupted download, or a malicious attack. "
                            "Delete the files at ~/.poli_objectives/rasp and try again."
                        )
                    else:
                        logging.warning(
                            f"The downloaded downstream model {path_.name} does not match the expected md5 checksum. "
                            "This could be due to a corrupted download, or a malicious attack. "
                        )
            elif verbose:
                print(
                    f"poli 🧪: Downstream model {path_.parent.name} already exists. Skipping."
                )

    def raw_pdb_to_unique_chain(self, wildtype_pdb_path: Path, chain: str = "A"):
        """
        This function takes a raw pdb file, extracts a chain, and tidies up
        using pdbtools. Following their interactive implementation, this
        function will:
            1. Select a chain.
            2. Delete heteroatoms.
            3. Delete residues between -999 and 0.
            4. Fix insertions.
            5. Tidy the PDB.

        The output of this function is a PDB file with a single chain,
        stored at the working directory of this RaspInterface instance.
        More precisely, it will be stored at
        working_dir / raw / {wildtype_pdb_path.stem}_{chain}.pdb
        """
        # First command to run (which uses pdbtools to select a chain,
        # delete heteroatoms, remove some residues, fix insertions,
        # and tidy the PDB).
        raw_output_path = (
            self.working_dir
            / "raw"
            / f"{wildtype_pdb_path.stem}_query_protein_uniquechain.pdb"
        )

        # We need to load and transform the pdb file into
        # a line buffer that pdbtools can understand.
        with open(wildtype_pdb_path, "r") as fp:
            all_lines = fp.readlines()

        # Step 1: selecting the chain
        selchain_result = pdb_selchain_run(all_lines, chain)

        # Step 2: deleting heteroatoms
        deleting_heteroatoms_result = pdb_delhetatm_run(selchain_result)

        # Step 3: deleting residues between -999 and 0
        deleting_residues_result = pdb_delres_run(
            deleting_heteroatoms_result, residue_range=(-999, 0), step=1
        )

        # Step 4: fixing insertions
        fix_inserts_result = pdb_fixinsert_run(deleting_residues_result, [])

        # Step 5: tidying up the pdb
        tidy_result = pdb_tidy_run(fix_inserts_result)

        # tidy_result is now a generator of lines, we can use
        # it to write the output file.
        with open(raw_output_path, "w") as fp:
            for line in tidy_result:
                fp.write(line)

    def unique_chain_to_clean_pdb(self, wildtype_pdb_path: Path):
        """
        This function takes a pdb with a single chain, and
        cleans it using RaSP's clean_pdb.py script.

        The output of this function is a cleaned pdb file
        stored at the working directory of this RaspInterface
        instance. More precisely, it will be stored at
        working_dir / cleaned / {wildtype_pdb_path.stem}_clean.pdb
        This stem is usually just {pdb_id}_{chain}.
        """
        clean_pdb(
            pdb_input_filename=str(
                self.working_dir
                / "raw"
                / f"{wildtype_pdb_path.stem}_query_protein_uniquechain.pdb"
            ),
            out_dir=str(self.working_dir / "cleaned"),
            reduce_executable=str(self.reduce_executable_path),
        )

    def cleaned_to_parsed_pdb(
        self,
        wildtype_pdb_path: Path,
        max_radius: float = 9.0,
        include_center: bool = False,
    ):
        """
        This function takes a cleaned pdb file, and extracts the
        residue environments using extract_environments.py.
        The output of this function is a .npz file in the
        parsed directory given by:
        working_dir / parsed / {wildtype_pdb_path.stem}.npz
        At this stage, this stem is usually {pdb_id}_{chain}_clean
        """
        # The command used by the original RaSP implementation:
        # !python3 /content/src/pdb_parser_scripts/extract_environments.py --pdb_in /content/data/test/predictions/cleaned/query_protein_uniquechain_clean.pdb  --out_dir /content/data/test/predictions/parsed/  #&> /dev/null

        # We modify the naming convention to make it easier. Now
        # the output file will be named:
        pdb_id = wildtype_pdb_path.stem
        extract_environments(
            pdb_filename=str(
                self.working_dir
                / "cleaned"
                / f"{wildtype_pdb_path.stem}_query_protein_uniquechain_clean.pdb"
            ),
            pdb_id=pdb_id,
            max_radius=max_radius,
            out_dir=str(self.working_dir / "parsed"),
            include_center=include_center,
        )

    def create_df_structure(
        self, wildtype_pdb_path: Path, mutant_residue_strings: List[str] = None
    ):
        """
        This function creates a pandas dataframe with the
        residue environments of a given pdb file, and
        prepares it for ALL single-site mutations.

        If a list of mutant residue strings is provided,
        we will only consider the mutations in that list.
        For example, if the wildtype_pdb_path has the
        following residue string:

        "MSEMETKQV"

        and the mutant_residue_strings is:

        ["ASEMETKQV", "RSEMETKQV", "NSEMETKQV"]

        then the output dataframe will only contain the
        mutations M1A, M1R, and M1N.

        If no mutant_residue_strings is provided, then
        we will consider ALL single-site mutations.
        """
        # TODO: this is the convention inside the cleaning scripts,
        # but it feels flimsy. We should probably change it.
        pdb_id = wildtype_pdb_path.stem.replace("_query_protein_uniquechain_clean", "")
        pdb_filename_for_df = (
            self.working_dir / "parsed" / f"{pdb_id}_coordinate_features.npz"
        )
        assert pdb_filename_for_df.exists(), (
            f"{pdb_filename_for_df} does not exist."
            " Remember to run the preprocessing step."
        )

        dataset_structure = ResidueEnvironmentsDataset(
            [str(pdb_filename_for_df)], transformer=None
        )

        resenv_dataset = {}
        for resenv in dataset_structure:
            key = f"--{'CUSTOM'}--{resenv.chain_id}--{resenv.pdb_residue_number}--{index_to_one(resenv.restype_index)}--"
            resenv_dataset[key] = resenv

        df_structure_no_mt = pd.DataFrame.from_dict(
            resenv_dataset, orient="index", columns=["resenv"]
        )
        df_structure_no_mt.reset_index(inplace=True)
        df_structure_no_mt["index"] = df_structure_no_mt["index"].astype(str)
        res_info = pd.DataFrame(
            df_structure_no_mt["index"].str.split("--").tolist(),
            columns=["blank", "pdb_id", "chain_id", "pos", "wt_AA", "blank2"],
        )

        df_structure_no_mt["pdbid"] = res_info["pdb_id"]
        df_structure_no_mt["chainid"] = res_info["chain_id"]
        df_structure_no_mt["variant"] = res_info["wt_AA"] + res_info["pos"] + "X"
        aa_list = self.alphabet

        # Considering ALL single mutations in ALL sites.
        wildtype_residue_string = "".join(
            [v[0] for v in df_structure_no_mt["variant"].values]
        )

        df_structure = pd.DataFrame(
            df_structure_no_mt.values.repeat(20, axis=0),
            columns=df_structure_no_mt.columns,
        )
        pos_of_variant_column = df_structure.columns.get_loc("variant")
        for i in range(0, len(df_structure), 20):
            for j in range(20):
                df_structure.iloc[i + j, pos_of_variant_column] = (
                    df_structure.iloc[i + j, :]["variant"][:-1] + aa_list[j]
                )

        # This is a silly and inefficient way of doing this.
        # We should instead only build df_structure to have
        # the relevant variants from the start. That way
        # we avoid the above two-for-loops. But they're
        # practically instantaneous, so it's not a big deal.
        # (O(nm) doesn't matter all that much if n < 1000
        # , m is always 20, and the cost of each operation
        # is negligible)
        df_structure["mutant_residue_string"] = [""] * len(df_structure)
        if mutant_residue_strings is not None:
            # Compute the mutations associated to all strings in
            # mutant_residue_strings.

            # First, the following function computes where
            # the mutations are between the wildtype and
            # the mutant residue strings.
            # These are in the format ("replace", index_in_wildtype, index_in_mutant)
            mutations_in_rasp_format = []
            for mutant_residue_string in mutant_residue_strings:
                if mutant_residue_string == wildtype_residue_string:
                    # Then we append a mock mutation.
                    original_residue_as_single_character = wildtype_residue_string[0]
                    position_in_chain = res_info.iloc[0]["pos"]
                    mutant_residue_as_single_character = wildtype_residue_string[0]

                    mutation_in_rasp_format = (
                        original_residue_as_single_character
                        + f"{position_in_chain}"
                        + mutant_residue_as_single_character
                    )

                    mutations_in_rasp_format.append(mutation_in_rasp_format)

                    mask_for_this_mutation = df_structure["variant"].str.startswith(
                        mutation_in_rasp_format
                    )

                    df_structure.loc[
                        mask_for_this_mutation, "mutant_residue_string"
                    ] = mutant_residue_string
                    continue

                edits_ = edits_between_strings(
                    wildtype_residue_string, mutant_residue_string
                )
                for edit_ in edits_:
                    _, i, _ = edit_
                    original_residue_as_single_character = wildtype_residue_string[i]
                    position_in_chain = res_info.iloc[i]["pos"]
                    mutant_residue_as_single_character = mutant_residue_string[i]
                    mutation_in_rasp_format = (
                        original_residue_as_single_character
                        + f"{position_in_chain}"
                        + mutant_residue_as_single_character
                    )

                    mutations_in_rasp_format.append(mutation_in_rasp_format)

                    # Add the mutant residue string to the dataframe.
                    mask_for_this_mutation = df_structure["variant"].str.startswith(
                        mutation_in_rasp_format
                    )

                    df_structure.loc[
                        mask_for_this_mutation, "mutant_residue_string"
                    ] = mutant_residue_string

            # Filter df_structure to only contain the mutations in
            # mutations_in_rasp_format. (i.e. we need to focus on
            # only some of the positions, which are in the middle
            # of each string in mutations_in_rasp_format).
            df_structure = df_structure[
                df_structure["variant"].str.startswith(
                    tuple(set(mutations_in_rasp_format))
                )
            ]

            # We should attach the original mutant string to each
            # row in df_structure.

        df_structure.drop(columns="index", inplace=True)

        # Load PDB amino acid frequencies used to approximate unfolded states
        THIS_DIR = Path(__file__).parent.resolve()
        pdb_nlfs = -np.log(np.load(THIS_DIR / "pdb_frequencies.npz")["frequencies"])

        # # Add wt and mt idxs and freqs to df

        df_structure["wt_idx"] = df_structure.apply(
            lambda row: one_to_index(row["variant"][0]), axis=1
        )
        df_structure["mt_idx"] = df_structure.apply(
            lambda row: one_to_index(row["variant"][-1]), axis=1
        )
        df_structure["wt_nlf"] = df_structure.apply(
            lambda row: pdb_nlfs[row["wt_idx"]], axis=1
        )
        df_structure["mt_nlf"] = df_structure.apply(
            lambda row: pdb_nlfs[row["mt_idx"]], axis=1
        )

        return df_structure



"""Aloha objective factory and function.

This is a registration script for the ALOHA problem,
a simple example of a discrete black box objective
function where the goal is to find the sequence
["A", "L", "O", "H", "A"] among all 5-letter sequences.

The problem is registered as 'aloha', and it uses
a conda environment called 'poli__base' (see the
environment.yml file in this folder).
"""

from string import ascii_uppercase

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.seeding import seed_python_numpy_and_torch


class AlohaBlackBox(AbstractBlackBox):
    """
    Black box implementation for the Aloha problem.

    The aloha problem is a simple discrete black box problem
    where the goal is to find the sequence ["A", "L", "O", "H", "A"]
    among all 5-letter sequences.

    Parameters
    ----------
    batch_size : int, optional
        The batch size for processing multiple inputs simultaneously, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize the computation, by default False.
    num_workers : int, optional
        The number of workers to use for parallel computation, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.

    Attributes
    ----------
    alphabet : dict
        The mapping of symbols to their corresponding indices in the alphabet.

    Methods
    -------
    _black_box(x, context=None)
        The main black box method that performs the computation, i.e.
        it computes the distance between the 5-letter sequence in
        x and the target sequence ["A", "L", "O", "H", "A"].
    """

    def __init__(
        self,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
    ):
        """
        Initialize the aloha black box object.

        Parameters
        ----------
        batch_size : int, optional
            The batch size for processing data, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize the processing, by default False.
        num_workers : int, optional
            The number of workers to use for parallel processing, by default None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        """
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="aloha",
            max_sequence_length=5,
            aligned=True,
            fixed_length=True,
            deterministic=True,
            alphabet=list(ascii_uppercase),
            log_transform_recommended=False,
            discrete=True,
            padding_token="",
        )

    # The only method you have to define
    def _black_box(self, x: np.ndarray, context: dict = None) -> np.ndarray:
        """
        Compute the distance of x to the sequence "ALOHA".

        Parameters
        ----------
        x : np.ndarray
            Input array of shape [b, 5] containing strings or integers.
            If the elements are integers, they are converted to strings using the alphabet.
        context : dict, optional
            Additional context information (default is None).

        Returns
        -------
        y: np.ndarray
            Array of shape [b, 1] containing the distances to the sequence "ALOHA".
        """
        assert x.dtype.kind == "U", "The input must be an array of strings."

        if x.shape[1] == 1:
            assert (
                len(set([len(x_i) for x_i in x])) == 1
            ), "All strings must have the same length."
            x = np.array([list(x_i[0]) for x_i in x])

        matches = x == np.array(["A", "L", "O", "H", "A"])
        values = np.sum(matches.reshape(x.shape[0], 5), axis=1, keepdims=True).reshape(
            x.shape[0], 1
        )
        return values


class AlohaProblemFactory(AbstractProblemFactory):
    """
    Factory for the Aloha problem.

    The aloha problem is a simple discrete black box problem
    where the goal is to find the sequence ["A", "L", "O", "H", "A"]
    among all 5-letter sequences.

    Methods
    -------
    create(...)
        Creates a problem instance with the specified parameters.
    """

    def create(
        self,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Returns an Aloha blackbox function and initial observations.

        Parameters
        -----------
        seed:  int, optional
            The seed for random number generation. Default is None.
        batch_size:  int, optional
            The batch size for parallel evaluation. Default is None.
        parallelize : bool, optional
            Flag indicating whether to parallelize the evaluation. Default is False.
        num_workers:  int, optional
            The number of workers for parallel evaluation. Default is None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        force_isolation: bool, optional
            Unused in this black box, since it runs directly on the base environment.

        Returns
        --------
        results: Tuple[AbstractBlackBox, np.ndarray, np.ndarray]:
            A tuple containing the blackbox function, initial observations
            for input variables, and initial observations for output
            variables.
        """
        # We set the seed for numpy and python
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        f = AlohaBlackBox(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        x0 = np.array([["A", "L", "O", "O", "F"]])

        aloha_problem = Problem(
            black_box=f,
            x0=x0,
        )

        return aloha_problem



"""Utilities for transforming levels to arrays and back."""

from itertools import product
from typing import List

import numpy as np


def level_to_list(level_txt: str) -> List[List[str]]:
    """
    Takes a level as a string and returns
    a list of lists of individual tokens.
    """
    # Returns a list by splitting the level text
    # by \n.
    as_list = level_txt.split("\n")
    return [list(row) for row in as_list if row != ""]


def level_to_array(level_txt: str) -> np.ndarray:
    """Parses a level from string to numpy array."""
    # Returns a np array by splitting the level text
    # by \n.
    return np.array(level_to_list(level_txt))


def levels_to_onehot(levels: np.ndarray, n_sprites: int = 11) -> np.ndarray:
    """Transforms an array [b, w, h] of integers into a one-hot array [b, n_sprites, w, h]."""
    batch_size, w, h = levels.shape
    y_onehot = np.zeros((batch_size, n_sprites, h, w))
    for b, level in enumerate(levels):
        # for loop through the batch, no?
        # print(level)
        for i, j in product(range(h), range(w)):
            c = level[i, j]
            y_onehot[b, c, i, j] = 1

    return y_onehot


def vectorized(prob_matrix, items):
    s = prob_matrix.cumsum(axis=0)
    r = np.random.rand(prob_matrix.shape[1])
    k = (s < r).sum(axis=0)
    return items[k]


def onehot_to_levels(levels_onehot: np.ndarray, sampling=False, seed=0) -> np.ndarray:
    """
    Transforms a level from probits to integers.
    """
    if sampling:
        # From log-softmax to softmax.
        levels_onehot = np.exp(levels_onehot)
        np.random.seed(seed)
        batch_size, n_classes, h, w = levels_onehot.shape
        # u = np.random.rand(n_classes)
        # U = np.zeros_like(levels_onehot)
        # for b in range(batch_size):
        #     for i, j in product(range(h), range(w)):
        #         U[b, :, i, j] = u
        # levels = (levels_onehot.cumsum(axis=1) > U).argmax(axis=1)

        # Is there a smarter way to do this?
        # There is:
        # https://stackoverflow.com/a/34190035/3516175
        levels = np.zeros((batch_size, h, w), dtype=int)
        for b in range(batch_size):
            for i, j in product(range(h), range(w)):
                p = levels_onehot[b, :, i, j]
                levels[b, i, j] = np.random.choice(n_classes, p=p)

        np.random.seed()
    else:
        levels = np.argmax(levels_onehot, axis=1)

    return levels


def add_padding_to_level(level: np.ndarray, n_padding: int = 1) -> np.ndarray:
    """
    Adds padding to the left of the level, giving room
    for the agent to land.
    """
    h, w = level.shape
    padding = 2 * np.ones((h, n_padding))  # Starting with emptyness.
    padding[-1, :] = 0  # Adding the ground.
    level_with_padding = np.concatenate((padding, level), axis=1)

    return level_with_padding


def clean_level(level: np.ndarray) -> List[List[int]]:
    """
    Cleans a level by removing Mario (token id: 11),
    and replacing it with empty space.
    """
    # Cleaning up Mario (11), replacing
    # it with empty space (2).
    level[level == 11] = 2
    level = level.astype(int)
    # level = add_padding_to_level(level, 2)

    return level.tolist()



"""Benchmarks for toy continuous objective functions

These benchmarks contain the usual toy continuous functions [1, 2].

References
----------

[1] Al-Roomi, A. R. (2015).
    Unconstrained Single-Objective Benchmark Functions Repository.
    Dalhousie University, Electrical and Computer Engineering. Halifax, Nova Scotia, Canada.
    Available at: https://www.al-roomi.org/benchmarks/unconstrained

[2] Surjanovic, S., Bingham, D., “Optimization Test Functions and Datasets.”
    Accessed April 12, 2024.
    https://www.sfu.ca/~ssurjano/optimization.html.
"""

from typing import List, Union

from poli.core.abstract_benchmark import AbstractBenchmark
from poli.core.problem import Problem
from poli.objective_repository import ToyContinuousProblemFactory
from poli.objective_repository.toy_continuous_problem.toy_continuous_problem import (
    POSSIBLE_FUNCTIONS,
    SIX_DIMENSIONAL_PROBLEMS,
    TWO_DIMENSIONAL_PROBLEMS,
)


class ToyContinuousFunctionsBenchmark(AbstractBenchmark):
    """
    A benchmark with several toy continuous functions.

    References
    ----------

    [1] Al-Roomi, A. R. (2015).
       Unconstrained Single-Objective Benchmark Functions Repository.
       Dalhousie University, Electrical and Computer Engineering. Halifax, Nova Scotia, Canada.
       Available at: https://www.al-roomi.org/benchmarks/unconstrained

    [2] Surjanovic, S., Bingham, D.,
        “Optimization Test Functions and Datasets.”
        Accessed April 12, 2024.
        https://www.sfu.ca/~ssurjano/optimization.html.
    """

    def __init__(
        self,
        n_dimensions: int = 2,
        embed_in: Union[int, None] = None,
        dimensions_to_embed_in: Union[List[int], None] = None,
        seed: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        parallelize: bool = False,
        num_workers: Union[int, None] = None,
        evaluation_budget: Union[int, List[int]] = float("inf"),
    ) -> None:
        super().__init__(
            seed=seed,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.n_dimensions = n_dimensions
        self.embed_in = embed_in
        self.dimensions_to_embed_in = dimensions_to_embed_in
        self.function_names = list(
            (
                set(POSSIBLE_FUNCTIONS)
                - set(TWO_DIMENSIONAL_PROBLEMS)
                - set(SIX_DIMENSIONAL_PROBLEMS)
            )
        )
        self.problem_factories = [ToyContinuousProblemFactory()] * len(
            self.function_names
        )

    def _initialize_problem(self, index: int) -> Problem:
        problem_factory: ToyContinuousProblemFactory = self.problem_factories[index]

        problem = problem_factory.create(
            function_name=self.function_names[index],
            n_dimensions=self.n_dimensions,
            embed_in=self.embed_in,
            dimensions_to_embed_in=self.dimensions_to_embed_in,
            seed=self.seed,
            batch_size=self.batch_size,
            parallelize=self.parallelize,
            num_workers=self.num_workers,
            evaluation_budget=self.evaluation_budget,
        )

        return problem


class EmbeddedBranin2D(AbstractBenchmark):
    """
    A benchmark with the Branin 2D function embedded in different dimensions

    In this benchmark, we provide the Branin 2D function embedded in
    5, 10, 25, 50, and 100 dimensions.

    References
    ----------

    [1] Al-Roomi, A. R. (2015).
       Unconstrained Single-Objective Benchmark Functions Repository.
       Dalhousie University, Electrical and Computer Engineering. Halifax, Nova Scotia, Canada.
       Available at: https://www.al-roomi.org/benchmarks/unconstrained
    [2] Surjanovic, S., Bingham, D.,
        “Optimization Test Functions and Datasets.”
        Accessed April 12, 2024.
        https://www.sfu.ca/~ssurjano/optimization.html.
    """

    def __init__(
        self,
        seed: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        parallelize: bool = False,
        num_workers: Union[int, None] = None,
        evaluation_budget: Union[int, List[int]] = float("inf"),
    ) -> None:
        super().__init__(
            seed,
            batch_size,
            parallelize,
            num_workers,
            evaluation_budget,
        )
        self.embed_in = [5, 10, 25, 50, 100]
        self.problem_factories = [ToyContinuousProblemFactory()] * len(self.embed_in)

    def _initialize_problem(self, index: int) -> Problem:
        problem_factory: ToyContinuousProblemFactory = self.problem_factories[index]

        problem = problem_factory.create(
            function_name="branin_2d",
            embed_in=self.embed_in[index],
            seed=self.seed,
            batch_size=self.batch_size,
            parallelize=self.parallelize,
            num_workers=self.num_workers,
            evaluation_budget=self.evaluation_budget,
        )

        return problem


class EmbeddedHartmann6D(AbstractBenchmark):
    """Benchmark with the Hartmann 6D function embedded in different dimensions

    In this benchmark, we provide the original Hartmann 6D function, plus
    its embedding into 10, 25, 50, and 100 dimensions.

    References
    ----------

    [1] Al-Roomi, A. R. (2015).
       Unconstrained Single-Objective Benchmark Functions Repository.
       Dalhousie University, Electrical and Computer Engineering. Halifax, Nova Scotia, Canada.
       Available at: https://www.al-roomi.org/benchmarks/unconstrained
    [2] Surjanovic, S., Bingham, D.,
        “Optimization Test Functions and Datasets.”
        Accessed April 12, 2024.
        https://www.sfu.ca/~ssurjano/optimization.html.
    """

    def __init__(
        self,
        seed: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        parallelize: bool = False,
        num_workers: Union[int, None] = None,
        evaluation_budget: Union[int, List[int]] = float("inf"),
    ) -> None:
        super().__init__(
            seed,
            batch_size,
            parallelize,
            num_workers,
            evaluation_budget,
        )
        self.embed_in = [None, 10, 25, 50, 100]
        self.problem_factories = [ToyContinuousProblemFactory()] * len(self.embed_in)

    def _initialize_problem(self, index: int) -> Problem:
        problem_factory: ToyContinuousProblemFactory = self.problem_factories[index]

        if index == 0:
            problem = problem_factory.create(
                function_name="hartmann_6d",
                n_dimensions=6,
                # embed_in=self.embed_in[index],
                seed=self.seed,
                batch_size=self.batch_size,
                parallelize=self.parallelize,
                num_workers=self.num_workers,
                evaluation_budget=self.evaluation_budget,
            )
        else:
            problem = problem_factory.create(
                function_name="hartmann_6d",
                embed_in=self.embed_in[index],
                seed=self.seed,
                batch_size=self.batch_size,
                parallelize=self.parallelize,
                num_workers=self.num_workers,
                evaluation_budget=self.evaluation_budget,
            )

        return problem



"""Utilities for defining and manipulating mutations on proteins.

This module contains utilities for defining
mutations on proteins according to foldx.

foldx expects mutations in a certain format:
    - the first letter is the original residue
    - the second letter is the position of the mutation
    - the third letter is the chain ID,
    - the fourth letter is the mutant residue.

See the "Individual List Mode" of https://foldxsuite.crg.eu/parameter/mutant-file
for more details.
"""

from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
from Bio.PDB.Residue import Residue
from Bio.SeqUtils import seq1

from poli.core.util.proteins.pdb_parsing import parse_pdb_as_residue_strings


def edits_between_strings(
    string_1: str, string_2: str, strict: bool = True
) -> List[Tuple[str, int, int]]:
    """
    Compute the edit operations between two strings.

    Parameters
    ----------
    string_1 : str
        The first string.
    string_2 : str
        The second string.
    strict : bool, optional
        If True, check if the lengths of string_1 and string_2 are equal.
        Defaults to True.

    Returns
    -------
    List[Tuple[str, int, int]]
        A list of tuples representing the edit operations between the two strings.
        Each tuple contains the operation type ("replace"), the position in string_1,
        and the position in string_2.

    Raises
    ------
    AssertionError
        If strict is True and the lengths of string_1 and string_2 are different.

    Examples
    --------
    >>> edits_between_strings("abc", "abd")
    [('replace', 2, 2)]

    >>> edits_between_strings("abc", "def")
    [('replace', 0, 0), ('replace', 1, 1), ('replace', 2, 2)]
    """
    if strict:
        assert len(string_1) == len(string_2), (
            f"string_1 and string_2 have different lengths: "
            f"{len(string_1)} and {len(string_2)}."
        )
    for i, (a, b) in enumerate(zip(string_1, string_2)):
        if a != b:
            yield ("replace", i, i)


def mutations_from_wildtype_residues_and_mutant(
    wildtype_residues: List[Residue], mutated_residue_string: str
) -> List[str]:
    """Computes the mutations from a wildtype list of residues
    and a mutated residue string.

    Since foldx expects an individual_list.txt file of mutations,
    this function computes the Levenshtein distance between
    the wildtype residue string and the mutated residue string,
    keeping track of the replacements.

    This method returns a list of strings which are to be written
    in a single line of individual_list.txt. Each string is a
    mutation in the format foldx expects (e.g. "EA1R", meaning
    that an E was mutated to an R in position 1 of chain A. The
    first letter is the original residue, the second letter is
    the chain, the third letter is the position, and the fourth
    letter is the mutant residue).

    If the mutated residue string is the same as the wildtype residue
    string, we still need to pass a dummy mutation to foldx, so we
    "mutate" the first residue in the wildtype string to itself.

    For example:
        wildtype_residue_string = "ECDE..."
        mutated_residue_string =  "ACDE..."

    This function would return (assuming that we are mutating the
    chain "A"):
        ["EA1A"]

    Parameters
    ----------
    wildtype_residues : List[Residue]
        The list of wildtype residues.
    mutated_residue_string : str
        The mutated residue string.

    Returns
    -------
    mutations: List[str]
        The list of mutations in the format foldx expects.
    """
    wildtype_residue_string = "".join(
        [seq1(res.get_resname()) for res in wildtype_residues]
    )

    # Making sure we treat the mutant string as uppercase
    mutated_residue_string = mutated_residue_string.upper()

    assert len(mutated_residue_string) == len(wildtype_residue_string), (
        f"wildtype residue string and mutated residue string "
        f"have different lengths: {len(wildtype_residue_string)} "
        f"and {len(mutated_residue_string)}."
    )

    # If the mutated string is the same as the wildtype string,
    # there are no mutations. Still, FoldX expects us to pass
    # a dummy mutation, so we "mutate" the first residue in the
    # wildtype string to itself.
    if mutated_residue_string == wildtype_residue_string:
        first_residue = wildtype_residues[0]
        first_residue_name = seq1(first_residue.get_resname())
        chain_id = first_residue.get_parent().id
        index_in_sequence = first_residue.id[1]
        return [
            f"{first_residue_name}{chain_id}{index_in_sequence}{first_residue_name}"
        ]

    mutations = edits_between_strings(wildtype_residue_string, mutated_residue_string)

    mutations_in_line = []
    for mutation_type, pos_in_wildtype, pos_in_mutant in mutations:
        assert mutation_type == "replace"
        original_wildtype_residue = wildtype_residue_string[pos_in_wildtype]
        chain_id = wildtype_residues[pos_in_wildtype].get_parent().id
        index_in_sequence = wildtype_residues[pos_in_wildtype].id[1]
        mutated_residue = mutated_residue_string[pos_in_mutant]

        mutation_string = (
            f"{original_wildtype_residue}{chain_id}{index_in_sequence}{mutated_residue}"
        )
        mutations_in_line.append(mutation_string)

    return mutations_in_line


def find_closest_wildtype_pdb_file_to_mutant(
    wildtype_pdb_files: List[Path],
    mutated_residue_string: str,
    return_hamming_distance: bool = False,
) -> Union[Path, Tuple[Path, int]]:
    """
    Find the closest wildtype PDB file to a given mutant residue string.

    Parameters
    ----------
    wildtype_pdb_files : List[Path]
        A list of paths to wildtype PDB files.
    mutated_residue_string : str
        The mutated residue string.
    return_hamming_distance : bool, optional
        If True, return the hamming distance along with the best candidate PDB file.
        Default is False.

    Returns
    -------
    Union[Path, Tuple[Path, int]]
        If return_hamming_distance is True, returns a tuple containing the best candidate PDB file
        and the hamming distance. Otherwise, returns the best candidate PDB file.

    Raises
    ------
    ValueError
        If no PDB file of the same length as the mutated residue string is found.

    """
    # First, we load up these pdb files as residue strings
    wildtype_residue_strings = {
        pdb_file: "".join(parse_pdb_as_residue_strings(pdb_file))
        for pdb_file in wildtype_pdb_files
    }

    # Since foldx only allows for substitutions, we only
    # consider the PDBs whose length is the same as the
    # mutated residue string.
    best_candidate_pdb_file = None
    min_hamming_distance = np.inf
    for pdb_file, wildtype_residue_string in wildtype_residue_strings.items():
        if len(wildtype_residue_string) != len(mutated_residue_string):
            continue

        hamming_distance = np.sum(
            [
                wildtype_residue_string.upper()[i] != mutated_residue_string.upper()[i]
                for i in range(len(wildtype_residue_string))
            ]
        )

        if hamming_distance < min_hamming_distance:
            best_candidate_pdb_file = pdb_file
            min_hamming_distance = hamming_distance

    if best_candidate_pdb_file is None:
        raise ValueError(
            f"Could not find a PDB file of length {len(mutated_residue_string)}." + "\n"
            "Are you sure you provided a valid mutant string?\n"
            f"Lengths allowed: {set([len(x) for x in wildtype_residue_strings.values()])}"
        )

    if return_hamming_distance:
        return best_candidate_pdb_file, min_hamming_distance
    else:
        return best_candidate_pdb_file



from pathlib import Path
from time import time
from typing import List, Union
from uuid import uuid4

import numpy as np

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.util.proteins.foldx import FoldxInterface
from poli.core.util.proteins.pdb_parsing import (
    parse_pdb_as_residue_strings,
    parse_pdb_as_residues,
)

# This is the folder where all the files
# generated by FoldX will be stored.
# Feel free to change it if you want
# to keep the files somewhere else by
# passing tmp_folder to the black box.
# TODO: what happens if the user is on Windows?
DEFAULT_TMP_PATH = Path("/tmp").resolve()


class FoldxIsolatedFunction(AbstractIsolatedFunction):
    """
    A class representing the FoldxBlackBox, which is used for simulating protein mutations using FoldX.

    Parameters
    -----------
    wildtype_pdb_path : Union[Path, List[Path]], required
        The path(s) to the wildtype PDB file(s). (default: None)
    experiment_id : str, optional
        The experiment ID. (default: None)
    tmp_folder : Path, optional
        The temporary folder path. (default: None)
    eager_repair : bool, optional
        Flag indicating whether to eagerly repair the PDB files. (default: False)
    verbose : bool, optional
        Flag indicating whether we print the output from FoldX. (default: False)

    Attributes
    ----------
    experiment_id : str
        The experiment ID.
    tmp_folder : Path
        The temporary folder path.
    wildtype_pdb_paths : List[Path]
        The list of repaired wildtype PDB file paths.
    wildtype_residues : List[List[Residue]]
        The list of wildtype residues for each PDB file.
    wildtype_amino_acids : List[List[str]]
        The list of wildtype amino acids for each PDB file.
    wildtype_residue_strings : List[str]
        The list of wildtype residue strings for each PDB file.

    Methods
    -------
    create_working_directory() -> Path:
        Creates and returns the working directory path for the black box.

    """

    def __init__(
        self,
        wildtype_pdb_path: Union[Path, List[Path]],
        experiment_id: str = None,
        tmp_folder: Path = None,
        eager_repair: bool = False,
        verbose: bool = False,
    ):
        """
        Initialize the FoldxBlackBox.

        Parameters
        -----------
        wildtype_pdb_path : Union[Path, List[Path]]
            The path(s) to the wildtype PDB file(s).
        experiment_id : str, optional
            The experiment ID. (default: None)
        tmp_folder : Path, optional
            The temporary folder path. (default: None)
        eager_repair : bool, optional
            Flag indicating whether to eagerly repair the PDB files. (default: False)
        verbose : bool, optional
            Flag indicating whether we print the output from FoldX. (default: False)
        """
        # TODO: assert that wildtype_pdb_file is provided
        assert wildtype_pdb_path is not None, (
            "Missing required argument wildtype_pdb_file. "
            "Did you forget to pass it to create and into the black box?"
        )
        self.verbose = verbose

        # Defining the experiment id
        if experiment_id is None:
            experiment_id = f"{int(time())}_{str(uuid4())[:8]}"
        self.experiment_id = experiment_id

        self.tmp_folder = tmp_folder if tmp_folder is not None else DEFAULT_TMP_PATH

        if isinstance(wildtype_pdb_path, str):
            wildtype_pdb_path = Path(wildtype_pdb_path.strip())

        if isinstance(wildtype_pdb_path, Path):
            wildtype_pdb_path = [wildtype_pdb_path]

        if isinstance(wildtype_pdb_path, list):
            _wildtype_pdb_path = []
            for pdb_file in wildtype_pdb_path:
                if isinstance(pdb_file, str):
                    pdb_file = Path(pdb_file.strip())
                assert isinstance(
                    pdb_file, Path
                ), f"Expected a Path object or a string, but got {type(pdb_file)}."
                _wildtype_pdb_path.append(pdb_file)

            wildtype_pdb_path = _wildtype_pdb_path

        # At this point, wildtype_pdb_path is a list of Path objects.
        # We need to ensure that these are repaired pdb files.
        # We do this by creating a temporary folder and repairing
        # the pdbs there.
        if eager_repair:
            path_for_repairing_pdbs = self.tmp_folder / "foldx_tmp_files_for_repair"
            path_for_repairing_pdbs.mkdir(exist_ok=True, parents=True)
            foldx_interface_for_repairing = FoldxInterface(
                path_for_repairing_pdbs, verbose=self.verbose
            )

            # Re-writing wildtype_pdb_path to be the list of repaired pdb files.
            repaired_wildtype_pdb_files = [
                foldx_interface_for_repairing._repair_if_necessary_and_provide_path(
                    pdb_file
                )
                for pdb_file in wildtype_pdb_path
            ]

            # At this point, wildtype_pdb_path is a list of Path objects.
            self.wildtype_pdb_paths = repaired_wildtype_pdb_files
        else:
            self.wildtype_pdb_paths = wildtype_pdb_path

        self.wildtype_resiudes = [
            parse_pdb_as_residues(pdb_file) for pdb_file in self.wildtype_pdb_paths
        ]

        self.wildtype_amino_acids = [
            parse_pdb_as_residue_strings(pdb_file)
            for pdb_file in self.wildtype_pdb_paths
        ]

        self.wildtype_residue_strings = [
            "".join(amino_acids) for amino_acids in self.wildtype_amino_acids
        ]
        self.x0 = np.array(self.wildtype_residue_strings)

    def create_working_directory(self) -> Path:
        """
        Create and return the working directory path for the black box.

        Returns
        --------
        Path
            The path to the working directory.
        """
        sub_experiment_id = str(uuid4())[:8]

        working_dir = (
            self.tmp_folder / "foldx_tmp_files" / self.experiment_id / sub_experiment_id
        )
        working_dir.mkdir(exist_ok=True, parents=True)

        return working_dir

    def __call__(self, x: np.ndarray, context=None) -> np.ndarray:
        raise NotImplementedError



import numpy as np


def test_multi_objective_instantiation():
    from poli.core.multi_objective_black_box import MultiObjectiveBlackBox
    from poli.objective_repository import AlohaBlackBox

    f_aloha = AlohaBlackBox()

    f = MultiObjectiveBlackBox(
        objective_functions=[f_aloha, f_aloha],
    )

    assert f.objective_functions == [f_aloha, f_aloha]

    x0 = np.array([["A", "B", "C", "D", "E"]])
    y0 = f(x0)

    assert y0.shape == (1, 2)


def test_negative_black_boxes():
    from poli.objective_repository import AlohaBlackBox

    f = AlohaBlackBox()
    g = -f

    x0 = np.array([["A", "B", "C", "D", "E"]])

    f0 = f(x0)
    g0 = g(x0)

    assert f0 == -g0



from .guacamol import GuacaMolGoalDirectedBenchmark
from .pmo import PMOBenchmark
from .toy_continuous_functions_benchmark import (
    EmbeddedBranin2D,
    EmbeddedHartmann6D,
    ToyContinuousFunctionsBenchmark,
)

__all__ = [
    "GuacaMolGoalDirectedBenchmark",
    "PMOBenchmark",
    "ToyContinuousFunctionsBenchmark",
    "EmbeddedBranin2D",
    "EmbeddedHartmann6D",
]



"""
This module contains coding for training a Gaussian Process Regression
model on the Sarkisyan (2016) data set.
Plus some utility constants.
"""

import warnings

warnings.filterwarnings("ignore")
from pathlib import Path

import numpy as np

from poli.objective_repository.gfp_cbas import BLOSUM


class SequenceGP(object):
    def __init__(
        self,
        load: bool = False,
        X_train: np.ndarray = None,
        y_train: np.ndarray = None,
        length_scale: float = 1,
        homo_noise: float = 0.1,
        load_path_prefix: Path = Path(__file__).parent
        / "assets"
        / "models"
        / "gp"
        / "gfp_gp",
        k_beta: float = 0.1,
        c: float = 1,
        d: float = 2,
    ):
        if load:
            self.load(path=load_path_prefix)
        else:
            assert X_train is not None and y_train is not None
            self.X_ = np.copy(X_train)
            self.y_ = np.copy(y_train).reshape((y_train.shape[0], 1))
            self.N_ = self.X_.shape[0]
            self.params_ = np.array([homo_noise, k_beta, c, d])
            self.K_ = None
            self.Kinv_ = None

    def _kernel(self, Xi: np.ndarray, Xj: np.ndarray) -> float:
        """
        BLOSUM based product kernel of discrete input sequences.
        Hyperparameters are applied
        """
        beta = self.params_[1]
        c = self.params_[2]
        d = self.params_[3]
        kij = np.prod(BLOSUM[Xi, Xj] ** beta)
        kii = np.prod(BLOSUM[Xi, Xi] ** beta)
        kjj = np.prod(BLOSUM[Xj, Xj] ** beta)
        k = kij / (np.sqrt(kii * kjj))
        k = np.exp(c * k)
        return k

    def _fill_K(self, print_every: int = 100) -> None:
        self.K_ = np.zeros((self.N_, self.N_))
        total = self.N_ * (self.N_ + 1) / 2
        m = 0
        homo_noise = self.params_[0]
        for i in range(self.N_):
            for j in range(i, self.N_):
                kij = self._kernel(self.X_[i], self.X_[j])
                if i == j:
                    kij += homo_noise
                self.K_[i, j] = kij
                self.K_[j, i] = kij

                m += 1
                if m % print_every == 0:
                    print("Number of K elements filled: %i / %i" % (m, total))

    def _invert_K(self):
        print("Inverting K...")
        self.Kinv_ = np.linalg.inv(self.K_)
        print("Done inverting K.")

    def build(self, print_every=100):
        self._fill_K(print_every=print_every)
        self._invert_K()

    def predict(
        self, Xstar: np.ndarray, print_every: int = None, predict_variance: bool = False
    ) -> np.ndarray:
        """
        GP posterior predictive, compute kernel values across all new Xstar and existing observations X_ .
        """
        M = len(Xstar)
        Kstar = np.zeros((M, self.N_))
        total = M * self.N_
        m = 0
        for i in range(M):
            for j in range(self.N_):
                kij = self._kernel(Xstar[i], self.X_[j])
                Kstar[i, j] = kij
                m += 1
                if print_every is not None:
                    if m % print_every == 0:
                        print("Number of Kstar elements filled: %i / %i" % (m, total))
        mu_star = np.matmul(Kstar, np.matmul(self.Kinv_, self.y_))
        return mu_star

    def save(self, path: Path):
        """
        Persist numpy arrays from object properties.
        """
        np.save(path.parent / (path.name + "X.npy"), self.X_)
        np.save(path.parent / (path.name + "y.npy"), self.y_)
        np.save(path.parent / (path.name + "K.npy"), self.K_)
        np.save(path.parent / (path.name + "Kinv.npy"), self.Kinv_)
        np.save(path.parent / (path.name + "params.npy"), self.params_)

    def load(self, path: Path):
        """
        Load persisted files. Sets object properties.
        """
        self.X_ = np.load(path.parent / (path.name + "X.npy"))
        self.y_ = np.load(path.parent / (path.name + "y.npy"))
        self.K_ = np.load(path.parent / (path.name + "K.npy"))
        self.Kinv_ = np.load(path.parent / (path.name + "Kinv.npy"))
        self.params_ = np.load(path.parent / (path.name + "params.npy"))
        self.N_ = self.X_.shape[0]



from pathlib import Path
from typing import Literal
from warnings import warn

import numpy as np
import pandas as pd
import torch

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.black_box_information import BlackBoxInformation
from poli.objective_repository.gfp_cbas.cbas_alphabet_preprocessing import (
    convert_aas_to_idx_array,
    one_hot_encode_aa_array,
)
from poli.objective_repository.gfp_cbas.cbas_wrapper import CBASVAEWrapper
from poli.objective_repository.gfp_cbas.gfp_gp import SequenceGP


class GFPCBasIsolatedLogic(AbstractIsolatedFunction):
    def __init__(
        self,
        problem_type: Literal["gp", "vae", "elbo"],
        info: BlackBoxInformation,
        n_starting_points: int = 1,
        seed: int = None,
        functional_only: bool = False,
        ignore_stops: bool = True,
        unique=True,
    ):
        gfp_path = Path(__file__).parent.resolve() / "assets"
        self.info = info
        self.vae = CBASVAEWrapper(AA=len(info.alphabet), L=info.max_sequence_length).vae
        self.seed = seed

        data_df = pd.read_csv(gfp_path / "gfp_data.csv")[
            ["medianBrightness", "std", "nucSequence", "aaSequence"]
        ]
        gfp_wt_seq = np.atleast_1d(
            np.loadtxt(gfp_path / "avGFP_reference_sequence.fa", skiprows=1, dtype=str)
        )[0]
        self.reference_entry = data_df[
            data_df.nucSequence.str.lower() == gfp_wt_seq.lower()
        ]
        data_df = data_df.drop(
            self.reference_entry.index
        )  # take out WT reference sequence
        if unique:
            data_df = data_df.drop_duplicates(subset="aaSequence")
        if self.seed:  # if random seed is provided, shuffle the data
            data_df = data_df.sample(frac=1, random_state=seed)
        if (
            functional_only
        ):  # if functional only setting, threshold by median Brightness
            idx = data_df.loc[
                (data_df["medianBrightness"] > data_df["medianBrightness"].mean())
            ].index
        else:
            idx = data_df.index
        data_df = data_df.loc[idx]
        if ignore_stops:  # ignore incorrect encodings
            idx = data_df.loc[~data_df["aaSequence"].str.contains("!")].index
        self.model = None
        # use ProblemSetupInfo name to specify black-box function
        if problem_type == "gp":
            gp_file_prefix = (
                Path(__file__).parent.resolve() / "assets" / "models" / "gp" / "gfp_gp"
            )
            self.model = SequenceGP(load=True, load_path_prefix=gp_file_prefix)
            self.f = self._seq_gp_predict
        elif problem_type == "elbo":
            self.model = self.vae.vae_
            self.f = self._elbo_predict
        elif problem_type == "vae":
            self.model = self.vae.encoder_
            self.f = self._vae_embedding
        else:
            raise NotImplementedError(
                f"Misspecified info: {info.name} does not contain [gp ; elbo ; vae]!"
            )
        self.data_df = data_df.loc[idx]

        x0 = np.array(list(self.reference_entry.aaSequence.values[0]))[
            np.newaxis, :
        ]  # WT reference sequence
        if (
            n_starting_points > 1
        ):  # take a random subset of available AA sequence at request
            _x0 = np.array(
                [list(s) for s in self.data_df.aaSequence[: n_starting_points - 1]]
            )
            x0 = np.concatenate([x0, _x0])
            assert x0.shape[0] == n_starting_points

        self.x0 = x0

    def _seq_gp_predict(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate Listgarten GP on label-encoded sequence input.
        Return negative pred.mean of GP (minimization problem)
        """
        assert self.model and self.model.__class__.__name__ == "SequenceGP"
        le_x = convert_aas_to_idx_array(x)
        return -self.model.predict(le_x)  # NOTE: predict negative mu

    def _elbo_predict(self, x: np.ndarray) -> np.ndarray:
        """
        One Hot encode and VAE evaluate given reference zero point.
        Calls Keras engine function evaluation to compute ELBO.
        """
        assert self.model and self.model.__class__.__name__ == "Functional"
        oh_x = one_hot_encode_aa_array(x)
        # model.evaluate takes two array inputs: input , [decoder_out, KLD ref.]
        # TODO: batched evaluation not working: returns one value only (not size of batch) -> iterating - fix this for speed-up.
        kld_reference_prior = np.zeros([1, oh_x.shape[-1]])
        model_evaluation = np.array(
            [
                self.model.evaluate(
                    _oh_x_seq[np.newaxis, :],
                    [_oh_x_seq[np.newaxis, :], kld_reference_prior],
                    batch_size=1,
                    verbose=0,
                )
                for _oh_x_seq in oh_x
            ]
        )
        # subselect ELBO as first column:
        return -model_evaluation[:, 0].reshape(-1, 1)  # NOTE: minimize ELBO as target

    def _vae_embedding(self, x: np.ndarray) -> np.ndarray:
        """
        One hot encode sequence and VAE embed
        """
        assert self.model and self.model.__class__.__name__ == "Functional"
        oh_x = one_hot_encode_aa_array(x)
        return self.model.predict(oh_x)[0]

    def __call__(self, x: np.array, context=None) -> np.ndarray:
        """
        x is encoded sequence return function value given problem name
        """
        with torch.no_grad():
            cbas_mu = self.f(x)
        return np.array(cbas_mu)

    def __iter__(self, *args, **kwargs):
        warn(f"{self.__class__.__name__} iteration invoked. Not implemented!")


if __name__ == "__main__":
    from poli.core.registry import register_isolated_function

    register_isolated_function(
        GFPCBasIsolatedLogic,
        name="gfp_cbas__isolated",
        conda_environment_name="poli__protein_cbas",
    )



"""This module contains utilities for registering problems and observers.
"""

import configparser
import warnings
from pathlib import Path
from typing import List, Type, Union

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.util.abstract_observer import AbstractObserver
from poli.core.util.objective_management.make_run_script import (
    make_isolated_function_script,
    make_observer_script,
)

# from poli.objective_repository import AVAILABLE_PROBLEM_FACTORIES, AVAILABLE_OBJECTIVES

_DEFAULT = "DEFAULT"
_OBSERVER = "observer"
_RUN_SCRIPT_LOCATION = "run_script_location"
_ISOLATED_FUNCTION_SCRIPT_LOCATION = "isolated_function_script_location"
DEFAULT_OBSERVER_NAME = ""
_DEFAULT_OBSERVER_RUN_SCRIPT = ""

HOME_DIR = Path.home().resolve()
(HOME_DIR / ".poli_objectives").mkdir(exist_ok=True)

config_file = str(HOME_DIR / ".poli_objectives" / "config.rc")
config = configparser.ConfigParser(defaults={_OBSERVER: _DEFAULT_OBSERVER_RUN_SCRIPT})
ls = config.read(config_file)


def register_observer(
    observer: Union[AbstractObserver, Type[AbstractObserver]],
    conda_environment_location: str = None,
    python_paths: List[str] = None,
    observer_name: str = None,
    set_as_default_observer: bool = True,
):
    """Defines an external observer to be run in a separate process.

    This function takes an observer, a conda environment, a list of python
    environments, and an observer name. With these, it creates a script that
    can be run to instantiate the observer in a separate process. If no
    observer name is passed, the observer is set as the default observer.

    After registering an observer using this function, the user can instantiate
    it by using the ExternalObserver class, passing the relevant observer name.

    Parameters
    ----------
    observer : AbstractObserver
        The observer to be registered.
    conda_environment_location : str
        The location of the conda environment to be used.
    python_paths : List[str]
        A list of paths to append to the python path of the run script.
    observer_name : str
        The name of the observer to be registered.
    set_as_default_observer: bool, optional
        Whether to register the observer as the default observer. By default, this argument is true.
    Notes
    -----
    The observer script MUST accept port and password as arguments.
    """
    if observer_name == DEFAULT_OBSERVER_NAME:
        raise RuntimeError(
            "You must not use the name of the default observer which is '%s'"
            % DEFAULT_OBSERVER_NAME
        )
    if isinstance(observer, type):
        non_instance_observer = observer
    else:
        non_instance_observer = observer.__class__
    if observer_name is None:
        observer_name = observer.__name__
    run_script_location = make_observer_script(
        non_instance_observer, conda_environment_location, python_paths
    )
    _set_observer_run_script(
        run_script_location,
        observer_name=observer_name,
        set_as_default_observer=set_as_default_observer,
    )


def _set_observer_run_script(
    script_file_name: str, observer_name: str, set_as_default_observer: bool = True
) -> None:
    """Sets a run_script to be called on observer instantiation.

    This function takes as input the location of a script, and an observer name.
    Using these, it sets the configuration. If no observer name is passed, the
    observer is set as the default observer.

    Parameters
    ----------
    script_file_name : str
        The location of the script to be run.
    observer_name : str
        The name of the observer to be registered.

    Notes
    -----
    The observer script MUST accept port and password as arguments.
    """
    if _OBSERVER not in config.sections():
        config.add_section(_OBSERVER)
    # VERY IMPORTANT: the observer script MUST accept port and password as arguments
    config[_OBSERVER][observer_name] = script_file_name
    if set_as_default_observer:
        config[_DEFAULT][_OBSERVER] = observer_name
    _write_config()


def remove_default_observer():
    """
    This functions resets the default observer to do nothing.
    """
    config[_DEFAULT][_OBSERVER] = _DEFAULT_OBSERVER_RUN_SCRIPT
    _write_config()


def register_isolated_function(
    isolated_function: Union[AbstractBlackBox, AbstractIsolatedFunction],
    name: str,
    conda_environment_name: Union[str, Path] = None,
    python_paths: List[str] = None,
    force: bool = True,
    **kwargs,
):
    if name == _OBSERVER:
        raise RuntimeError(
            "The name " + name + " is a reserved keyword. Please choose another name."
        )
    if "conda_environment_location" in kwargs:
        conda_environment_name = kwargs["conda_environment_location"]

    if not name.endswith("__isolated"):
        warnings.warn(
            "By convention, the name of the isolated function should end with '__isolated'. "
            "This is to distinguish it from the original function. "
        )

    if name not in config.sections():
        config.add_section(name)
    elif not force:
        # If force is false, we warn the user and ask for confirmation
        user_input = input(
            f"Black box {name} has already been registered. "
            f"Do you want to overwrite it? (y/[n]) "
        )
        if user_input.lower() != "y":
            warnings.warn(f"Black box {name} already exists. Not overwriting.")
            return

        warnings.warn(f"Black box {name} already exists. Overwriting.")

    isolated_function_script_location = make_isolated_function_script(
        isolated_function, conda_environment_name, python_paths, **kwargs
    )
    config[name][_ISOLATED_FUNCTION_SCRIPT_LOCATION] = isolated_function_script_location
    _write_config()


def _write_config():
    with open(config_file, "w+") as configfile:
        config.write(configfile)



"""The exact same computation of LogP used by LamBO.

In [1], the authors optimize jointly for the Quantitative Estimate of
Druglikeness (QED) and a penalized version of log-solubility (logP).
For fair comparisons, we implement the exact same computation of logP.

References
----------
[1] “Accelerating Bayesian Optimization for Biological Sequence Design with Denoising Autoencoders.”
Stanton, Samuel, Wesley Maddox, Nate Gruver, Phillip Maffettone,
Emily Delaney, Peyton Greenside, and Andrew Gordon Wilson.
arXiv, July 12, 2022. http://arxiv.org/abs/2203.12742.
"""



"""Utilities for seeding random number generators."""

import random

import numpy as np


def seed_numpy(seed: int = None) -> None:
    """
    Seed the NumPy random number generator.

    Parameters
    ----------
    seed : int, optional
        Seed value for the random number generator. If None, then no seeding is performed.
    """
    if seed is not None:
        np.random.seed(seed)


def seed_python(seed: int = None) -> None:
    """
    Seed the random number generator for Python.


    Parameters
    ----------
    seed : int, optional
        Seed value for the random number generator. If None, then no seeding is performed.
    """
    if seed is not None:
        random.seed(seed)


def seed_torch(seed: int = None) -> None:
    """
    Seed the random number generator for PyTorch.


    Parameters
    ----------
    seed : int, optional
        Seed value for the random number generator. If None, no seeding is performed.
    """
    try:
        import torch
    except ImportError:
        return

    if seed is not None:
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)


def seed_python_numpy_and_torch(seed: int = None) -> None:
    """
    Seed all random number generators.


    Parameters
    ----------
    seed : int, optional
        Seed value for the random number generator. If None, no seeding is performed.
    """
    seed_numpy(seed)
    seed_python(seed)
    seed_torch(seed)



"""Registers the logP objective factory and black-box

This is a registration script for the rdkit_logp problem,
whose black box objective function returns the log quotient
of solubility (a.k.a. logP).

This black box is a simple wrapper around RDKit's
descriptors. We allow for both SMILES and SELFIES
strings.
"""

from __future__ import annotations

from typing import Literal

import numpy as np
from rdkit.Chem import Descriptors

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.chemistry.tdc_problem import TDCProblem
from poli.core.problem import Problem
from poli.core.util.chemistry.string_to_molecule import strings_to_molecules
from poli.core.util.seeding import seed_python_numpy_and_torch


class LogPBlackBox(AbstractBlackBox):
    """Log-solubility of a small molecule.

    A simple black box that returns the LogP
    of a molecule. By default, we assume that the
    result of concatenating the tokens will be
    a SMILES string, but you can set the context
    variable "from_selfies" to True to indicate
    that the input is a SELFIES string.

    RDKit's Chem.MolFromSmiles function and logP are known
    for failing silently, so we return NaN if the
    molecule cannot be parsed or if qed returns
    something other than a float.

    Parameters
    ----------
    string_representation : Literal["SMILES", "SELFIES"], optional
        The string representation to use, by default "SMILES".
    alphabet : list[str] | None, optional
        The alphabet to be used for the SMILES or SELFIES representation.
        It is common that the alphabet depends on the dataset used, so
        it is recommended to pass it as an argument. Default is None.
    max_sequence_length : int, optional
        The maximum length of the sequence. Default is infinity.
    batch_size : int, optional
        The batch size for processing multiple inputs simultaneously, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize the computation, by default False.
    num_workers : int, optional
        The number of workers to use for parallel computation, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.

    Attributes
    ----------
    from_selfies : bool
        Flag indicating whether the input is a SELFIES string.
    from_smiles : bool
        Flag indicating whether the input is a SMILES string.

    Methods
    -------
    _black_box(x, context=None)
        The main black box method that performs the computation, i.e.
        it computes the logP of the molecule in x.
    """

    def __init__(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        """
        Initializes the LogP black box.

        Parameters
        ----------
        string_representation : Literal["SMILES", "SELFIES"], optional
            The string representation to use, by default "SMILES".
        alphabet : list[str] | None, optional
            The alphabet to be used for the SMILES or SELFIES representation.
            It is common that the alphabet depends on the dataset used, so
            it is recommended to pass it as an argument. Default is None.
        max_sequence_length : int, optional
            The maximum length of the sequence. Default is infinity.
        batch_size : int, optional
            The batch size for processing multiple inputs simultaneously, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize the computation, by default False.
        num_workers : int, optional
            The number of workers to use for parallel computation, by default None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        """
        assert string_representation.upper() in ["SMILES", "SELFIES"]
        self.from_selfies = string_representation.upper() == "SELFIES"
        self.from_smiles = string_representation.upper() == "SMILES"
        self.alphabet = alphabet
        self.max_sequence_length = max_sequence_length
        self.string_representation = string_representation

        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

    # The only method you have to define
    def _black_box(self, x: np.ndarray, context: dict = None) -> np.ndarray:
        """Computes the logP of a molecule x (array of strings).

        Assuming that x is an array of integers of length L,
        we use the alphabet to construct a SMILES string,
        and query logP from RDKit.
        """
        if x.dtype.kind in ["U", "S"]:
            molecule_strings = ["".join([x_ij for x_ij in x_i.flatten()]) for x_i in x]
        else:
            raise ValueError(
                f"Unsupported dtype: {x.dtype}. "
                "The input must be an array of strings."
            )

        # Transforms strings into RDKit molecules.
        # Those that cannot be parsed are set to None.
        molecules = strings_to_molecules(
            molecule_strings, from_selfies=self.from_selfies
        )

        # Computes the LogP values for each molecule.
        logp_values = []

        for molecule in molecules:
            if molecule is not None:
                logp_value = Descriptors.MolLogP(molecule)

                # If the qed value is not a float, return NaN
                if not isinstance(logp_value, float):
                    logp_value = np.nan

            # If the molecule is None, then RDKit failed
            # to parse it, and we return NaN.
            else:
                logp_value = np.nan

            logp_values.append(logp_value)

        return np.array(logp_values).reshape(-1, 1)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="rdkit_logp",
            max_sequence_length=self.max_sequence_length,
            aligned=False,
            fixed_length=False,
            deterministic=True,
            alphabet=self.alphabet,  # TODO: add once we settle for one
            log_transform_recommended=False,
            discrete=True,
            fidelity=None,
            padding_token="",
        )


class LogPProblemFactory(AbstractProblemFactory):
    def create(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """Creates a logP problem instance.

        Parameters
        ----------
        string_representation : str, optional
            The string representation of the input, by default "SMILES".
            Supported values are "SMILES" and "SELFIES".
        seed : int, optional
            The seed value for random number generation, by default None.
        batch_size : int, optional
            The batch size for processing multiple inputs simultaneously, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize the computation, by default False.
        num_workers : int, optional
            The number of workers to use for parallel computation, by default None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.

        Returns
        -------
        problem : Problem
            The logP problem instance, containing the black box and the initial input.
        """
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        if string_representation.upper() not in ["SMILES", "SELFIES"]:
            raise ValueError(
                "Missing required keyword argument: string_representation: str. "
                "String representation must be either 'SMILES' or 'SELFIES'."
            )

        self.string_representation = string_representation

        f = LogPBlackBox(
            string_representation=string_representation.upper(),
            alphabet=alphabet,
            max_sequence_length=max_sequence_length,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

        # The sequence "C"
        if string_representation.upper() == "SMILES":
            x0 = np.array([["C" * 10]])
        else:
            x0 = np.array([["[C]" * 10]])

        problem = TDCProblem(f, x0)

        return problem



"""
This test module tests whether the seeding works
in the white noise example.
"""

import numpy as np

from poli import objective_factory


def test_seeding_in_white_noise_factory_creation():
    problem_1 = objective_factory.create(
        "white_noise", seed=42, batch_size=1, parallelize=False
    )
    f, x0 = problem_1.black_box, problem_1.x0
    y0 = f(x0)

    problem_2 = objective_factory.create(
        "white_noise", seed=42, batch_size=1, parallelize=False
    )
    f, x0 = problem_2.black_box, problem_2.x0
    y1 = f(x0)

    problem_3 = objective_factory.create(
        "white_noise", seed=43, batch_size=1, parallelize=False
    )
    f, x0 = problem_3.black_box, problem_3.x0
    y2 = f(x0)

    assert (y0 == y1).all() and not np.isclose(y0, y2).all()
